import logging

import rich.syntax
import textual.binding
import textual.widgets

import borgmatic.actions.browse.loading
import borgmatic.actions.browse.workers

logger = logging.getLogger('__name__')


class File_preview(textual.widgets.RichLog):
    '''
    A widget for extracting and previewing the contents of a file stored in a Borg archive.
    '''

    BINDINGS = (
        *textual.widgets.RichLog.BINDINGS,
        textual.binding.Binding(
            key='up,k', action='scroll_up', description='scroll up', show=True, priority=True
        ),
        textual.binding.Binding(
            key='down,j', action='scroll_down', description='scroll down', show=True, priority=True
        ),
        textual.binding.Binding(
            key='pageup', action='page_up', description='page up', show=True, priority=True
        ),
        textual.binding.Binding(
            key='pagedown', action='page_down', description='page down', show=True, priority=True
        ),
    )

    def __init__(self, config, repository, archive_name, file_path):
        '''
        Given a configuration dict, a repository dict, an archive name, and the path of a file in
        the archive, prepare to load the file's contents for eventual display in this widget. Actual
        loading kicks off in on_mount() below.
        '''
        self.config = config
        self.repository = repository
        self.archive_name = archive_name
        self.file_path = file_path

        super().__init__(classes='panel')
        self.border_title = f'📄 {self.file_path} preview'
        self.auto_scroll = False
        self.file_preview_loaded = borgmatic.actions.browse.workers.File_preview_loaded(
            self, 'file preview loaded'
        )

        self.loading_timer = borgmatic.actions.browse.loading.add_inline_loading_indicator(self)

    def on_mount(self):
        '''
        When this widget gets mounted in the DOM, subscribe to archive loaded events so that we can
        find out about archives as they load. Also start loading file contents from the archive.

        Loading is started *after* subscribing to the file preview loaded signal so that there's not
        a gap where we might miss out on signal publishes.
        '''
        self.file_preview_loaded.subscribe(self, self.on_file_preview_loaded)

        borgmatic.actions.browse.workers.load_file_preview(
            self.app,
            file_preview_loaded=self.file_preview_loaded,
            config=self.config,
            repository=self.repository,
            archive_name=self.archive_name,
            file_path=self.file_path,
        )

    def on_file_preview_loaded(self, file_contents):
        '''
        When a file loads, write its contents (syntax highlighted) to this file preview widget.
        '''
        self.loading_timer.stop()
        self.clear()

        if file_contents is None:
            self.write('Cannot display a preview for this file')
        else:
            # Only pass the file path and not its contents to guess_lexer(). Passing the contents is
            # more accurate, but also much slower.
            self.write(
                rich.syntax.Syntax(file_contents, rich.syntax.Syntax.guess_lexer(self.file_path))
            )



import logging
import os
import shlex
import subprocess

import borgmatic.config.paths
import borgmatic.config.validate
from borgmatic.borg import environment, feature, flags, repo_list
from borgmatic.execute import DO_NOT_CAPTURE, execute_command

logger = logging.getLogger(__name__)


def extract_last_archive_dry_run(
    config,
    local_borg_version,
    global_arguments,
    repository_path,
    lock_wait=None,
    local_path='borg',
    remote_path=None,
):
    '''
    Perform an extraction dry-run of the most recent archive. If there are no archives, skip the
    dry-run.
    '''
    extra_borg_options = config.get('extra_borg_options', {}).get('extract', '')
    verbosity_flags = ()
    if logger.isEnabledFor(logging.DEBUG):
        verbosity_flags = ('--debug', '--show-rc')
    elif logger.isEnabledFor(logging.INFO):
        verbosity_flags = ('--info',)

    try:
        last_archive_name = repo_list.resolve_archive_name(
            repository_path,
            'latest',
            config,
            local_borg_version,
            global_arguments,
            local_path,
            remote_path,
        )
    except ValueError:
        logger.warning('No archives found. Skipping extract consistency check.')
        return

    list_flag = ('--list',) if logger.isEnabledFor(logging.DEBUG) else ()
    full_extract_command = (
        (local_path, 'extract', '--dry-run')
        + (('--remote-path', remote_path) if remote_path else ())
        + (('--log-json',) if not config.get('progress') else ())
        + (('--lock-wait', str(lock_wait)) if lock_wait else ())
        + verbosity_flags
        + (('--progress',) if config.get('progress') else ())
        + list_flag
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + flags.make_repository_archive_flags(
            repository_path,
            last_archive_name,
            local_borg_version,
        )
    )

    if config.get('progress'):
        execute_command(
            full_extract_command,
            output_file=DO_NOT_CAPTURE,
            environment=environment.make_environment(config),
            working_directory=borgmatic.config.paths.get_working_directory(config),
            borg_local_path=local_path,
            borg_exit_codes=config.get('borg_exit_codes'),
        )
    else:
        execute_command(
            full_extract_command,
            environment=environment.make_environment(config),
            working_directory=borgmatic.config.paths.get_working_directory(config),
            borg_local_path=local_path,
            borg_exit_codes=config.get('borg_exit_codes'),
        )


def extract_archive(
    dry_run,
    repository,
    archive,
    paths,
    config,
    local_borg_version,
    global_arguments,
    local_path='borg',
    remote_path=None,
    destination_path=None,
    strip_components=None,
    extract_to_stdout=False,
):
    '''
    Given a dry-run flag, a local or remote repository path, an archive name, zero or more paths to
    restore from the archive, the local Borg version string, an argparse.Namespace of global
    arguments, a configuration dict, optional local and remote Borg paths, and an optional
    destination path to extract to, extract the archive into the current directory.

    If extract to stdout is True, then start the extraction streaming to stdout, and return that
    extract process as an instance of subprocess.Popen.
    '''
    umask = config.get('umask')
    lock_wait = config.get('lock_wait')
    extra_borg_options = config.get('extra_borg_options', {}).get('extract', '')

    if feature.available(feature.Feature.NUMERIC_IDS, local_borg_version):
        numeric_ids_flags = ('--numeric-ids',) if config.get('numeric_ids') else ()
    else:
        numeric_ids_flags = ('--numeric-owner',) if config.get('numeric_ids') else ()

    if strip_components == 'all':
        if not paths:
            raise ValueError('The --strip-components flag with "all" requires at least one --path')

        # Calculate the maximum number of leading path components of the given paths. "if piece"
        # ignores empty path components, e.g. those resulting from a leading slash. And the "- 1"
        # is so this doesn't count the final path component, e.g. the filename itself.
        strip_components = max(
            0,
            *(
                len(tuple(piece for piece in path.split(os.path.sep) if piece)) - 1
                for path in paths
            ),
        )

    working_directory = borgmatic.config.paths.get_working_directory(config)

    full_command = (
        (local_path, 'extract')
        + (('--remote-path', remote_path) if remote_path else ())
        + numeric_ids_flags
        + (('--umask', str(umask)) if umask else ())
        + (('--log-json',) if (config.get('log_json') or not config.get('progress')) else ())
        + (('--lock-wait', str(lock_wait)) if lock_wait else ())
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug', '--list', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + (('--dry-run',) if dry_run else ())
        + (('--strip-components', str(strip_components)) if strip_components else ())
        + (('--progress',) if config.get('progress') and not extract_to_stdout else ())
        + (('--stdout',) if extract_to_stdout else ())
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + flags.make_repository_archive_flags(
            # Make the repository path absolute so the destination directory used below via changing
            # the working directory doesn't prevent Borg from finding the repo. But also apply the
            # user's configured working directory (if any) to the repo path.
            borgmatic.config.validate.normalize_repository_path(repository, working_directory),
            archive,
            local_borg_version,
        )
        + (tuple(paths) if paths else ())
    )

    borg_exit_codes = config.get('borg_exit_codes')
    full_destination_path = (
        os.path.join(working_directory or '', destination_path) if destination_path else None
    )

    if extract_to_stdout:
        return execute_command(
            full_command,
            output_file=subprocess.PIPE,
            run_to_completion=False,
            environment=environment.make_environment(config),
            working_directory=full_destination_path,
            borg_local_path=local_path,
            borg_exit_codes=borg_exit_codes,
        )

    # The progress output isn't compatible with captured and logged output, as progress messes with
    # the terminal directly.
    if config.get('progress'):
        return execute_command(
            full_command,
            output_file=DO_NOT_CAPTURE,
            environment=environment.make_environment(config),
            working_directory=full_destination_path,
            borg_local_path=local_path,
            borg_exit_codes=borg_exit_codes,
        )
        return None

    # Don't give Borg local path so as to error on warnings, as "borg extract" only gives a warning
    # if the restore paths don't exist in the archive.
    execute_command(
        full_command,
        environment=environment.make_environment(config),
        working_directory=full_destination_path,
        borg_local_path=local_path,
        borg_exit_codes=borg_exit_codes,
    )

    return None



import logging

import borgmatic.actions.pattern
import borgmatic.borg.diff

logger = logging.getLogger(__name__)


def run_diff(
    repository,
    config,
    local_borg_version,
    diff_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "diff" action for the given repository.
    '''

    # Only process patterns if only_patterns flag is set
    if diff_arguments.only_patterns:
        working_directory = borgmatic.config.paths.get_working_directory(config)
        processed_patterns = borgmatic.actions.pattern.process_patterns(
            (*borgmatic.actions.pattern.collect_patterns(config, working_directory),),
            config,
            working_directory,
        )
    else:
        processed_patterns = None

    archive = borgmatic.borg.repo_list.resolve_archive_name(
        repository['path'],
        diff_arguments.archive,
        config,
        local_borg_version,
        global_arguments,
        local_path,
        remote_path,
    )
    second_archive = borgmatic.borg.repo_list.resolve_archive_name(
        repository['path'],
        diff_arguments.second_archive,
        config,
        local_borg_version,
        global_arguments,
        local_path,
        remote_path,
    )

    borgmatic.borg.diff.diff(
        repository['path'],
        archive,
        second_archive,
        config,
        local_borg_version,
        diff_arguments,
        global_arguments,
        local_path=local_path,
        remote_path=remote_path,
        patterns=processed_patterns,
    )



import os

import textual.widgets

import borgmatic.actions.browse.bindings


class Configuration_files_list(textual.widgets.OptionList):
    '''
    A widget for selecting a single borgmatic configuration file from among available configuration
    files. The item selection event is handled in a Carousel instance, the parent widget of an
    Configuration_files_list.
    '''

    BINDINGS = borgmatic.actions.browse.bindings.OPTION_LIST_BINDINGS

    def __init__(self, configs):
        '''
        Given a dict mapping from configuration path to corresponding configuration dict, add each
        configuration path as an option to this widget.
        '''
        self.configs = configs
        home_directory = os.path.expanduser('~')

        super().__init__(
            *(
                textual.widgets.option_list.Option(unexpanded_path, id=config_path)
                for config_path in configs
                for unexpanded_path in (config_path.replace(home_directory, '~'),)
            ),
            classes='panel',
        )
        self.border_title = '📄 configuration files'



import os

import textual.widgets

import borgmatic.actions.browse.bindings
import borgmatic.actions.browse.icons
import borgmatic.actions.browse.loading
import borgmatic.actions.browse.workers


def get_relative_archive_path_components(archive_path, current_directory_path_components):
    '''
    Given an Archive_path instance and a tuple of path components for the currently browsed
    directory, get the path components as a tuple for the archive path relative to that directory.

    For instance, given an archive path with a path of 'foo/bar/baz/quux.txt' and current
    directory path components of ('foo', 'bar'), return ('baz', 'quux.txt').

    If the archive path is not actually relative to the current directory, return None.
    '''
    archive_path_components = tuple(archive_path.file_path.split(os.path.sep))

    if not current_directory_path_components:
        return archive_path_components

    # If the loaded path doesn't match this directory list's own path, then we don't care about
    # it for purposes of displaying this particular directory.
    if (
        tuple(archive_path_components[: len(current_directory_path_components)])
        != current_directory_path_components
    ):
        return None

    # Strip off the portion of the archive path that matches the directory list's own path.
    return archive_path_components[len(current_directory_path_components) :]


def make_directory_list_option(archive_path, relative_path_components):
    '''
    Given an Archive_path instance and a tuple of relative path components for it, make a
    textual.widgets.option_list.Option for the path. Use an the icon based on whether this looks
    like a terminal filename or a directory.
    '''
    pieces = (
        borgmatic.actions.browse.icons.PATH_TYPE_ICONS.get(
            archive_path.path_type if len(relative_path_components) == 1 else 'd', '❓'
        ),
        relative_path_components[0],
    ) + (('→', archive_path.link_target) if archive_path.link_target else ())

    return textual.widgets.option_list.Option(
        prompt=' '.join(pieces), id=relative_path_components[0]
    )


def add_archive_paths(
    directory_list,
    config,
    repository,
    archive_name,
    archive_paths,
):
    '''
    Given a DirectoryList instance, a configuration dict, a repository dict, an archive name, and a
    sequence of ArchivePath instances, add the paths to the directory list as options, sorting and
    deduplicating the resulting directory list's options.

    After all of this reshuffling, make sure the original highlighted option remains highlighted.
    '''
    highlighted_option = directory_list.highlighted_option
    original_options_count = len(directory_list.options)

    sorted_options = sorted(
        (
            *directory_list.options,
            *(
                make_directory_list_option(archive_path, relative_path_components)
                for archive_path in archive_paths
                for relative_path_components in (
                    get_relative_archive_path_components(
                        archive_path,
                        directory_list.path_components,
                    ),
                )
                if relative_path_components
                if relative_path_components[0] not in directory_list._id_to_option
            ),
        ),
        # The loading indicator "option" always goes to the bottom.
        key=lambda option: ((option.id == 'loading-indicator'), option.prompt),
    )

    # If there aren't actually any options to add (due to deduplication), bail.
    if len(sorted_options) == original_options_count:
        return

    # Retain the highlighted option position even as other options load around it.
    directory_list.set_options(sorted_options)
    directory_list.highlighted = (
        directory_list.get_option_index(highlighted_option.id)
        if highlighted_option and directory_list.highlighted_option_changed
        else 0
    )


class Directory_list(textual.widgets.OptionList):
    '''
    A widget for selecting a path from among the contents of a particular directory in a Borg
    archive. The item selection event is handled in a Carousel instance, the parent widget of a
    Directory_list.
    '''

    BINDINGS = borgmatic.actions.browse.bindings.OPTION_LIST_BINDINGS

    def __init__(self, config, repository, archive_name, path_loaded=None, path_components=None):
        '''
        Given a configuration dict, a repository dict, an archive name, an optional
        Archive_path_loaded instance for signalling new paths as they load, and an optional tuple of
        path components indicating this directory's position in the backed up filesystem, prepare to
        load paths from the archive for eventual display in this widget. Actual loading kicks off in
        on_mount() below.
        '''
        self.config = config
        self.repository = repository
        self.archive_name = archive_name
        self.path_components = path_components or ()
        self.highlighted_option_changed = False

        super().__init__(classes='panel')

        self.border_title = ' '.join(
            (
                '📁',
                os.path.sep.join(self.path_components)
                if self.path_components
                else f'{archive_name}',
            )
        )

        if self.path_components:
            self.add_option(
                textual.widgets.option_list.Option(
                    '📁 ..',
                    id='..',
                ),
            )

        self.path_loaded = path_loaded or borgmatic.actions.browse.workers.Archive_path_loaded(
            self, 'archive path loaded'
        )

        if not self.path_loaded.complete:
            self.timer = borgmatic.actions.browse.loading.add_inline_loading_indicator(self)

    def on_mount(self):
        '''
        When this widget gets mounted in the DOM, subscribe to path loaded events so that we can
        find out about relevant archive paths as they load. And if this is a root directory list,
        start loading paths from the archive. If this is a non-root directory list, add any already
        loaded archive paths to this widget as options.

        Loading is started *after* subscribing to path loaded signals so that there's not a gap
        where we might miss out on any paths.
        '''
        self.path_loaded.subscribe(self, self.on_archive_path_loaded)

        if self.path_components:
            add_archive_paths(
                directory_list=self,
                config=self.config,
                repository=self.repository,
                archive_name=self.archive_name,
                archive_paths=borgmatic.actions.browse.workers.get_paths(
                    self.path_loaded.path_hierarchy, self.path_components
                ),
            )
        else:
            borgmatic.actions.browse.workers.load_archive_paths(
                self.app,
                path_loaded=self.path_loaded,
                config=self.config,
                repository=self.repository,
                archive_name=self.archive_name,
            )

    def on_archive_path_loaded(self, data):
        '''
        When an archive path loads, add it as an option to this directory list. But if we get a
        signal that all path loading is complete, stop and remove our loading indicator.
        '''
        if data is borgmatic.actions.browse.workers.LOADING_DONE:
            self.timer.stop()
            self.remove_option('loading-indicator')
            return

        add_archive_paths(
            directory_list=self,
            config=self.config,
            repository=self.repository,
            archive_name=self.archive_name,
            archive_paths=(data,),
        )

    def on_option_list_option_highlighted(self, event):
        '''
        When the highlighted option changes, record that fact. This flag is consumed in
        add_archive_paths() in order to retain the highlighted option even as other options load
        around it.
        '''
        if self.highlighted not in {None, 0}:
            self.highlighted_option_changed = True



import argparse
import json
import logging
import shlex

import borgmatic.config.paths
from borgmatic.borg import environment, feature, flags, repo_info
from borgmatic.execute import DO_NOT_CAPTURE, execute_command

logger = logging.getLogger(__name__)


def make_archive_filter_flags(local_borg_version, config, checks, check_arguments):
    '''
    Given the local Borg version, a configuration dict, a parsed sequence of checks, and check
    arguments as an argparse.Namespace instance, transform the checks into tuple of command-line
    flags for filtering archives in a check command.

    If "check_last" is set in the configuration and "archives" is in checks, then include a "--last"
    flag. And if "prefix" is set in configuration and "archives" is in checks, then include a
    "--match-archives" flag.
    '''
    check_last = config.get('check_last')
    prefix = config.get('prefix')

    if 'archives' in checks or 'data' in checks:
        return (('--last', str(check_last)) if check_last else ()) + (
            (
                ('--match-archives', f'sh:{prefix}*')
                if feature.available(feature.Feature.MATCH_ARCHIVES, local_borg_version)
                else ('--glob-archives', f'{prefix}*')
            )
            if prefix
            else (
                flags.make_match_archives_flags(
                    config.get('match_archives'),
                    config.get('archive_name_format'),
                    local_borg_version,
                )
            )
        )

    if check_last:
        logger.warning(
            'Ignoring check_last option, as "archives" or "data" are not in consistency checks',
        )

    if prefix:
        logger.warning(
            'Ignoring consistency prefix option, as "archives" or "data" are not in consistency checks',
        )

    return ()


def make_check_name_flags(checks, archive_filter_flags):
    '''
    Given parsed checks set and a sequence of flags to filter archives, transform the checks into
    tuple of command-line check flags.

    For example, given parsed checks of:

        ('repository',)

    This will be returned as:

        ('--repository-only',)

    However, if both "repository" and "archives" are in checks, then omit the "only" flags from the
    returned flags because Borg does both checks by default. Note that a "data" check only works
    along with an "archives" check.
    '''
    data_flags = ('--verify-data',) if 'data' in checks else ()
    common_flags = (archive_filter_flags if 'archives' in checks else ()) + data_flags

    if {'repository', 'archives'}.issubset(checks):
        return common_flags

    return (
        tuple(f'--{check}-only' for check in checks if check in {'repository', 'archives'})
        + common_flags
    )


def get_repository_id(
    repository_path,
    config,
    local_borg_version,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Given a local or remote repository path, a configuration dict, the local Borg version, global
    arguments, and local/remote commands to run, return the corresponding Borg repository ID.

    Raise ValueError if the Borg repository ID cannot be determined.
    '''
    try:
        return json.loads(
            repo_info.display_repository_info(
                repository_path,
                config,
                local_borg_version,
                argparse.Namespace(json=True),
                global_arguments,
                local_path,
                remote_path,
            ),
        )['repository']['id']
    except (json.JSONDecodeError, KeyError):
        raise ValueError(f'Cannot determine Borg repository ID for {repository_path}')


def check_archives(
    repository_path,
    config,
    local_borg_version,
    check_arguments,
    global_arguments,
    checks,
    archive_filter_flags,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a local or remote repository path, a configuration dict, the local Borg version, check
    arguments as an argparse.Namespace instance, global arguments, a set of named Borg checks to run
    (some combination "repository", "archives", and/or "data"), archive filter flags, and
    local/remote commands to run, check the contained Borg archives for consistency.
    '''
    lock_wait = config.get('lock_wait')
    extra_borg_options = config.get('extra_borg_options', {}).get('check', '')

    verbosity_flags = ()

    if logger.isEnabledFor(logging.INFO):
        verbosity_flags = ('--info',)

    if logger.isEnabledFor(logging.DEBUG):
        verbosity_flags = ('--debug', '--show-rc')

    try:
        repository_check_config = next(
            check for check in config.get('checks', ()) if check.get('name') == 'repository'
        )
    except StopIteration:
        repository_check_config = {}

    max_duration = check_arguments.max_duration or repository_check_config.get('max_duration')

    # If not configured, elevate Borg's exit code 1 (an ostensible warning) to error, because Borg
    # returns exit code 1 for repository check errors!
    borg_exit_codes = [*config.get('borg_exit_codes', []), *[{'code': 1, 'treat_as': 'error'}]]
    umask = config.get('umask')
    working_directory = borgmatic.config.paths.get_working_directory(config)

    if 'data' in checks:
        checks.add('archives')

    grouped_checks = (checks,)

    # If max_duration is set, then archives and repository checks need to be run separately, as Borg
    # doesn't support --max-duration along with an archives checks.
    if max_duration and 'archives' in checks and 'repository' in checks:
        checks.remove('repository')
        grouped_checks = (checks, {'repository'})

    for checks_subset in grouped_checks:
        full_command = (
            (local_path, 'check')
            + (('--repair',) if check_arguments.repair else ())
            + (
                ('--max-duration', str(max_duration))
                if max_duration and 'repository' in checks_subset
                else ()
            )
            + make_check_name_flags(checks_subset, archive_filter_flags)
            + (('--remote-path', remote_path) if remote_path else ())
            + (('--umask', str(umask)) if umask else ())
            + (
                ('--log-json',)
                if (
                    config.get('log_json') or not (check_arguments.repair or config.get('progress'))
                )
                else ()
            )
            + (('--lock-wait', str(lock_wait)) if lock_wait else ())
            + verbosity_flags
            + (('--progress',) if config.get('progress') else ())
            + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
            + flags.make_repository_flags(repository_path, local_borg_version)
        )

        execute_command(
            full_command,
            # The Borg repair option triggers an interactive prompt, which won't work when output is
            # captured. And progress messes with the terminal directly.
            output_file=(
                DO_NOT_CAPTURE if check_arguments.repair or config.get('progress') else None
            ),
            environment=environment.make_environment(config),
            working_directory=working_directory,
            borg_local_path=local_path,
            borg_exit_codes=borg_exit_codes,
        )



import contextlib
import logging

IS_A_HOOK = False
PAYLOAD_TRUNCATION_INDICATOR = '...\n'


class Forgetful_buffering_handler(logging.Handler):
    '''
    A buffering log handler that stores log messages in memory, and throws away messages (oldest
    first) once a particular capacity in bytes is reached. But if the given byte capacity is zero,
    don't throw away any messages.

    The given identifier is used to distinguish the instance of this handler used for one monitoring
    hook from those instances used for other monitoring hooks.
    '''

    def __init__(self, identifier, byte_capacity, log_level):
        super().__init__()

        self.identifier = identifier
        self.byte_capacity = byte_capacity
        self.byte_count = 0
        self.buffer = []
        self.forgot = False
        self.setLevel(log_level)

    def emit(self, record):
        message = record.getMessage() + '\n'
        self.byte_count += len(message)
        self.buffer.append(message)

        if not self.byte_capacity:
            return

        while self.byte_count > self.byte_capacity and self.buffer:
            self.byte_count -= len(self.buffer[0])
            self.buffer.pop(0)
            self.forgot = True


def add_handler(handler):
    '''
    Add the given handler to the global logger.
    '''
    logger = logging.getLogger()
    logger.addHandler(handler)
    logger.setLevel(min(handler.level for handler in logger.handlers))


def get_handler(identifier):
    '''
    Given the identifier for an existing Forgetful_buffering_handler instance, return the handler.

    Raise ValueError if the handler isn't found.
    '''
    try:
        return next(
            handler
            for handler in logging.getLogger().handlers
            if isinstance(handler, Forgetful_buffering_handler) and handler.identifier == identifier
        )
    except StopIteration:
        raise ValueError(f'A buffering handler for {identifier} was not found')


def format_buffered_logs_for_payload(identifier):
    '''
    Get the handler previously added to the root logger, and slurp buffered logs out of it to
    send to the monitoring service.
    '''
    try:
        buffering_handler = get_handler(identifier)
    except ValueError:
        # No handler means no payload.
        return ''

    payload = ''.join(message for message in buffering_handler.buffer)

    if buffering_handler.forgot:
        return PAYLOAD_TRUNCATION_INDICATOR + payload

    return payload


def remove_handler(identifier):
    '''
    Given the identifier for an existing Forgetful_buffering_handler instance, remove it.
    '''
    logger = logging.getLogger()

    with contextlib.suppress(ValueError):
        logger.removeHandler(get_handler(identifier))

    logger.setLevel(min(handler.level for handler in logger.handlers))



import json
import logging
import os
import platform
import time

import requests

from borgmatic.hooks.monitoring import monitor

logger = logging.getLogger(__name__)

MONITOR_STATE_TO_LOKI = {
    monitor.State.START: 'Started',
    monitor.State.FINISH: 'Finished',
    monitor.State.FAIL: 'Failed',
}
TIMEOUT_SECONDS = 10

# Threshold at which logs get flushed to loki
MAX_BUFFER_LINES = 100


class Loki_log_buffer:
    '''
    A log buffer that allows to output the logs as Loki requests in json. Allows
    adding labels to the log stream and takes care of communication with Loki.
    '''

    def __init__(self, url, dry_run, tls_cert_path=None, tls_key_path=None):
        '''
        Given a Loki URL, a dry run flag, and optional TLS certificate and key paths for mTLS authentication,
        create an instance of Loki_log_buffer.
        '''

        self.url = url
        self.dry_run = dry_run
        self.tls_cert_path = tls_cert_path
        self.tls_key_path = tls_key_path
        self.root = {'streams': [{'stream': {}, 'values': []}]}

    def add_value(self, value):
        '''
        Add a log entry to the stream.
        '''
        timestamp = str(time.time_ns())
        self.root['streams'][0]['values'].append((timestamp, value))

    def add_label(self, label, value):
        '''
        Add a label to the logging stream.
        '''
        self.root['streams'][0]['stream'][label] = value

    def to_request(self):
        return json.dumps(self.root)

    def __len__(self):
        '''
        Gets the number of lines currently in the buffer.
        '''
        return len(self.root['streams'][0]['values'])

    def flush(self):
        if self.dry_run:
            # Just empty the buffer and skip
            self.root['streams'][0]['values'] = []
            logger.info('Skipped uploading logs to Loki due to dry run')
            return

        if len(self) == 0:
            # Skip as there are not logs to send yet
            return

        request_body = self.to_request()
        self.root['streams'][0]['values'] = []

        try:
            result = requests.post(
                self.url,
                data=request_body,
                timeout=TIMEOUT_SECONDS,
                headers={
                    'Content-Type': 'application/json',
                    'User-Agent': 'borgmatic',
                },
                cert=(self.tls_cert_path, self.tls_key_path) if self.tls_cert_path else None,
            )
            result.raise_for_status()
        except requests.RequestException:
            logger.warning('Failed to upload logs to Loki')


class Loki_log_handler(logging.Handler):
    '''
    A log handler that sends logs to Loki.
    '''

    def __init__(self, url, send_logs, log_level, dry_run, tls_cert_path=None, tls_key_path=None):
        '''
        Given a URL to send logs to, whether all borgmatic logs should be sent (or just explicitly
        added messages from this hook), the log level to use (influencing which logs get sent), and
        whether this is a dry run, create an instance of Loki_log_buffer.
        '''
        super().__init__()

        self.buffer = Loki_log_buffer(
            url, dry_run, tls_cert_path=tls_cert_path, tls_key_path=tls_key_path
        )
        self.send_logs = send_logs
        self.setLevel(log_level)

    def emit(self, record):
        '''
        Add a general log record from the logging module to the stream—but only if send logs is
        enabled.
        '''
        if self.send_logs:
            self.raw(record.getMessage())

    def add_label(self, key, value):
        '''
        Add a label to the logging stream.
        '''
        self.buffer.add_label(key, value)

    def raw(self, msg):
        '''
        Add an arbitrary string as a log entry to the stream.
        '''
        self.buffer.add_value(msg)

        # If log sending is enabled, flush the buffer (and send data to Loki) once we accumulate
        # enough log data in the buffer. But if log sending is disabled, flush immediately so that,
        # for instance, start backup notifications are sent when the backup starts instead of after
        # it finishes!
        if len(self.buffer) > MAX_BUFFER_LINES or not self.send_logs:
            self.buffer.flush()

    def flush(self):
        '''
        Send the logs to Loki and empty the buffer.
        '''
        self.buffer.flush()


def initialize_monitor(hook_config, config, config_filename, monitoring_log_level, dry_run):
    '''
    Add a handler to the root logger to regularly send the logs to Loki.
    '''
    url = hook_config.get('url')
    tls = hook_config.get('tls', {})

    if bool(tls.get('cert_path')) != bool(tls.get('key_path')):
        raise ValueError(
            'Invalid Loki TLS configuration: cert_path and key_path must both be set or both be unset'
        )

    loki = Loki_log_handler(
        url,
        hook_config.get('send_logs', False),
        monitoring_log_level,
        dry_run,
        tls_cert_path=tls.get('cert_path'),
        tls_key_path=tls.get('key_path'),
    )

    for key, value in hook_config.get('labels').items():
        if value == '__hostname':
            loki.add_label(key, platform.node())
        elif value == '__config':
            loki.add_label(key, os.path.basename(config_filename))
        elif value == '__config_path':
            loki.add_label(key, config_filename)
        else:
            loki.add_label(key, value)

    global_logger = logging.getLogger()
    global_logger.addHandler(loki)
    global_logger.setLevel(min(handler.level for handler in global_logger.handlers))


def ping_monitor(hook_config, config, config_filename, state, monitoring_log_level, dry_run):
    '''
    Add an entry to the Loki logger with the current state.
    '''
    for handler in tuple(logging.getLogger().handlers):
        if isinstance(handler, Loki_log_handler) and state in MONITOR_STATE_TO_LOKI:
            handler.raw(f'{MONITOR_STATE_TO_LOKI[state]} backup')


def destroy_monitor(hook_config, config, monitoring_log_level, dry_run):
    '''
    Remove the monitor handler that was added to the root logger.
    '''
    global_logger = logging.getLogger()

    for handler in tuple(global_logger.handlers):
        if isinstance(handler, Loki_log_handler):
            handler.flush()
            global_logger.removeHandler(handler)

    global_logger.setLevel(min(handler.level for handler in global_logger.handlers))



import shlex
from argparse import Action
from textwrap import dedent

import borgmatic.commands.arguments
import borgmatic.commands.completion.actions
import borgmatic.config.validate


def has_file_options(action: Action):
    '''
    Given an argparse.Action instance, return True if it takes a file argument.
    '''
    return action.metavar in {
        'FILENAME',
        'PATH',
    } or action.dest in {'config_paths'}


def has_choice_options(action: Action):
    '''
    Given an argparse.Action instance, return True if it takes one of a predefined set of arguments.
    '''
    return action.choices is not None


def has_unknown_required_param_options(action: Action):
    '''
    A catch-all for options that take a required parameter, but we don't know what the parameter is.
    This should be used last. These are actions that take something like a glob, a list of numbers,
    or a string.

    Actions that match this pattern should not show the normal arguments, because those are unlikely
    to be valid.
    '''
    return (
        action.required is True
        or action.nargs
        in {
            '+',
            '*',
        }
        or action.metavar in {'PATTERN', 'KEYS', 'N'}
        or (action.type is not None and action.default is None)
    )


def has_exact_options(action: Action):
    return (
        has_file_options(action)
        or has_choice_options(action)
        or has_unknown_required_param_options(action)
    )


def exact_options_completion(action: Action):
    '''
    Given an argparse.Action instance, return a completion invocation that forces file completions,
    options completion, or just that some value follow the action, if the action takes such an
    argument and was the last action on the command line prior to the cursor.

    Otherwise, return an empty string.
    '''

    if not has_exact_options(action):
        return ''

    args = ' '.join(action.option_strings)

    if has_file_options(action):
        return f'''\ncomplete -c borgmatic -Fr -n "__borgmatic_current_arg {args}"'''

    if has_choice_options(action):
        return f'''\ncomplete -c borgmatic -f -a '{' '.join(map(str, action.choices))}' -n "__borgmatic_current_arg {args}"'''

    if has_unknown_required_param_options(action):
        return f'''\ncomplete -c borgmatic -x -n "__borgmatic_current_arg {args}"'''

    raise ValueError(
        f'Unexpected action: {action} passes has_exact_options but has no choices produced',
    )


def dedent_strip_as_tuple(string: str):
    '''
    Dedent a string, then strip it to avoid requiring your first line to have content, then return a
    tuple of the string. Makes it easier to write multiline strings for completions when you join
    them with a tuple.
    '''
    return (dedent(string).strip('\n'),)


def fish_completion():
    '''
    Return a fish completion script for the borgmatic command. Produce this by introspecting
    borgmatic's command-line argument parsers.
    '''
    (
        _,
        action_parsers,
        global_plus_action_parser,
    ) = borgmatic.commands.arguments.make_parsers(
        schema=borgmatic.config.validate.load_schema(borgmatic.config.validate.schema_filename()),
        unparsed_arguments=(),
    )

    all_action_parsers = ' '.join(action for action in action_parsers.choices)

    exact_option_args = tuple(
        ' '.join(action.option_strings)
        for action_parser in action_parsers.choices.values()
        for action in action_parser._actions
        if has_exact_options(action)
    ) + tuple(
        ' '.join(action.option_strings)
        for action in global_plus_action_parser._actions
        if len(action.option_strings) > 0
        if has_exact_options(action)
    )

    # Avert your eyes.
    # fmt: off
    return '\n'.join((
        *dedent_strip_as_tuple(
            f'''
            function __borgmatic_check_version
                set -fx this_filename (status current-filename)
                fish -c '
                    if test -f "$this_filename"
                        set this_script (cat $this_filename 2> /dev/null)
                        set installed_script (borgmatic --fish-completion 2> /dev/null)
                        if [ "$this_script" != "$installed_script" ] && [ "$installed_script" != "" ]
                            echo "{borgmatic.commands.completion.actions.upgrade_message(
                            'fish',
                            'borgmatic --fish-completion | sudo tee $this_filename',
                            '$this_filename',
                        )}"
                        end
                    end
                ' &
            end
            __borgmatic_check_version

            function __borgmatic_current_arg --description 'Check if any of the given arguments are the last on the command line before the cursor'
                set -l all_args (commandline -poc)
                # premature optimization to avoid iterating all args if there aren't enough
                # to have a last arg beyond borgmatic
                if [ (count $all_args) -lt 2 ]
                    return 1
                end
                for arg in $argv
                    if [ "$arg" = "$all_args[-1]" ]
                        return 0
                    end
                end
                return 1
            end

            set --local action_parser_condition "not __fish_seen_subcommand_from {all_action_parsers}"
            set --local exact_option_condition "not __borgmatic_current_arg {' '.join(exact_option_args)}"
            ''',
        ),
        '\n# action_parser completions',
        *tuple(
            f'''complete -c borgmatic -f -n "$action_parser_condition" -n "$exact_option_condition" -a '{action_name}' -d {shlex.quote(action_parser.description)}'''
            for action_name, action_parser in action_parsers.choices.items()
        ),
        '\n# global flags',
        *tuple(
            # -n is checked in order, so put faster / more likely to be true checks first
            f'''complete -c borgmatic -f -n "$exact_option_condition" -a '{' '.join(action.option_strings)}' -d {shlex.quote(action.help)}{exact_options_completion(action)}'''
            for action in global_plus_action_parser._actions
            # ignore the noargs action, as this is an impossible completion for fish
            if len(action.option_strings) > 0
            if 'Deprecated' not in action.help
        ),
        '\n# action_parser flags',
        *tuple(
            f'''complete -c borgmatic -f -n "$exact_option_condition" -a '{' '.join(action.option_strings)}' -d {shlex.quote(action.help)} -n "__fish_seen_subcommand_from {action_name}"{exact_options_completion(action)}'''
            for action_name, action_parser in action_parsers.choices.items()
            for action in action_parser._actions
            if 'Deprecated' not in (action.help or ())
        ),
    ))



import logging
import os
import shlex

import borgmatic.execute

logger = logging.getLogger(__name__)


SECRET_SERVICE_DATABASE_PATH = 'secret-service'


def load_credential(hook_config, config, credential_parameters):
    '''
    Given the hook configuration dict, the configuration dict, and a credential parameters tuple
    containing a KeePassXC database path and an attribute name to load, run keepassxc-cli to fetch
    the corresponding KeePassXC credential and return it. Or use secret-tool if the database path
    is "secret-service", indicating that KeePassXC's secret service integration should be used
    instead.

    Raise ValueError if keepassxc-cli or secret-tool can't retrieve the credential.
    '''
    try:
        (database_path, attribute_name) = credential_parameters
    except ValueError:
        raise ValueError(f'Invalid KeePassXC credential parameters: {credential_parameters}')

    if database_path == SECRET_SERVICE_DATABASE_PATH:
        command = (
            *shlex.split((hook_config or {}).get('secret_tool_command', 'secret-tool')),
            *('lookup', 'Path', attribute_name),
        )

        return '\n'.join(borgmatic.execute.execute_command_and_capture_output(command)).rstrip(
            os.linesep
        )

    expanded_database_path = os.path.expanduser(database_path)

    if not os.path.exists(expanded_database_path):
        raise ValueError(f'KeePassXC database path does not exist: {database_path}')

    # Build the keepassxc-cli command.
    command = (
        tuple(shlex.split((hook_config or {}).get('keepassxc_cli_command', 'keepassxc-cli')))
        + ('show', '--show-protected', '--attributes', 'Password')
        + (('--no-password',) if not (hook_config or {}).get('ask_for_password', True) else ())
        + (
            ('--key-file', hook_config['key_file'])
            if hook_config and hook_config.get('key_file')
            else ()
        )
        + (
            ('--yubikey', hook_config['yubikey'])
            if hook_config and hook_config.get('yubikey')
            else ()
        )
        + (expanded_database_path, attribute_name)  # Ensure database and entry are last.
    )

    return '\n'.join(borgmatic.execute.execute_command_and_capture_output(command)).rstrip(
        os.linesep
    )



import logging
import shlex

import borgmatic.borg.environment
import borgmatic.borg.feature
import borgmatic.borg.flags
import borgmatic.config.paths
import borgmatic.execute

logger = logging.getLogger(__name__)


FORCE_HARDER_FLAG_COUNT = 2


def make_repo_delete_command(
    repository,
    config,
    local_borg_version,
    repo_delete_arguments,
    global_arguments,
    local_path,
    remote_path,
    output_file,
):
    '''
    Given a local or remote repository dict, a configuration dict, the local Borg version, the
    arguments to the repo_delete action as an argparse.Namespace, and global arguments, the Borg
    local path, the Borg remote path, and an optional output file, return a command as a tuple to
    repo_delete the entire repository.
    '''
    extra_borg_options = config.get('extra_borg_options', {}).get(
        'repo_delete'
        if borgmatic.borg.feature.available(
            borgmatic.borg.feature.Feature.REPO_DELETE, local_borg_version
        )
        else 'delete',
        '',
    )

    return (
        (local_path,)
        + (
            ('repo-delete',)
            if borgmatic.borg.feature.available(
                borgmatic.borg.feature.Feature.REPO_DELETE,
                local_borg_version,
            )
            else ('delete',)
        )
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + borgmatic.borg.flags.make_flags('dry-run', global_arguments.dry_run)
        + borgmatic.borg.flags.make_flags('remote-path', remote_path)
        + borgmatic.borg.flags.make_flags('umask', config.get('umask'))
        + (('--log-json',) if output_file is None else ())
        + borgmatic.borg.flags.make_flags('lock-wait', config.get('lock_wait'))
        + borgmatic.borg.flags.make_flags('list', config.get('list_details'))
        + (
            (
                ('--force',)
                + (('--force',) if repo_delete_arguments.force >= FORCE_HARDER_FLAG_COUNT else ())
            )
            if repo_delete_arguments.force
            else ()
        )
        + borgmatic.borg.flags.make_flags_from_arguments(
            repo_delete_arguments,
            excludes=('list_details', 'force', 'repository'),
        )
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + borgmatic.borg.flags.make_repository_flags(repository['path'], local_borg_version)
    )


def delete_repository(
    repository,
    config,
    local_borg_version,
    repo_delete_arguments,
    global_arguments,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a local or remote repository dict, a configuration dict, the local Borg version, the
    arguments to the repo_delete action as an argparse.Namespace, global arguments as an
    argparse.Namespace, and local and remote Borg paths, repo_delete the entire repository.
    '''
    borgmatic.logger.add_custom_log_levels()

    # Don't capture output when Borg is expected to prompt for interactive confirmation, or the
    # prompt won't work.
    output_file = (
        None
        if repo_delete_arguments.force or repo_delete_arguments.cache_only
        else borgmatic.execute.DO_NOT_CAPTURE
    )

    command = make_repo_delete_command(
        repository,
        config,
        local_borg_version,
        repo_delete_arguments,
        global_arguments,
        local_path,
        remote_path,
        output_file,
    )

    borgmatic.execute.execute_command(
        command,
        output_log_level=logging.ANSWER,
        output_file=output_file,
        environment=borgmatic.borg.environment.make_environment(config),
        working_directory=borgmatic.config.paths.get_working_directory(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )



import logging
import os
import shlex

import borgmatic.config.paths
import borgmatic.logger
from borgmatic.borg import environment, flags
from borgmatic.execute import DO_NOT_CAPTURE, execute_command

logger = logging.getLogger(__name__)


def export_key(
    repository_path,
    config,
    local_borg_version,
    export_arguments,
    global_arguments,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a local or remote repository path, a configuration dict, the local Borg version, export
    arguments, and optional local and remote Borg paths, export the repository key to the
    destination path indicated in the export arguments.

    If the destination path is empty or "-", then print the key to stdout instead of to a file.

    Raise FileExistsError if a path is given but it already exists on disk.
    '''
    borgmatic.logger.add_custom_log_levels()
    umask = config.get('umask')
    lock_wait = config.get('lock_wait')
    working_directory = borgmatic.config.paths.get_working_directory(config)
    extra_borg_options = config.get('extra_borg_options', {}).get('key_export', '')

    if export_arguments.path and export_arguments.path != '-':
        if os.path.exists(os.path.join(working_directory or '', export_arguments.path)):
            raise FileExistsError(
                f'Destination path {export_arguments.path} already exists. Aborting.',
            )

        output_file = None
    else:
        output_file = DO_NOT_CAPTURE

    full_command = (
        (local_path, 'key', 'export')
        + (('--remote-path', remote_path) if remote_path else ())
        + (('--umask', str(umask)) if umask else ())
        + (('--log-json',) if output_file is None else ())
        + (('--lock-wait', str(lock_wait)) if lock_wait else ())
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + flags.make_flags('paper', export_arguments.paper)
        + flags.make_flags('qr-html', export_arguments.qr_html)
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + flags.make_repository_flags(
            repository_path,
            local_borg_version,
        )
        + ((export_arguments.path,) if output_file is None else ())
    )

    if global_arguments.dry_run:
        logger.info('Skipping key export (dry run)')
        return

    execute_command(
        full_command,
        output_file=output_file,
        output_log_level=logging.ANSWER,
        environment=environment.make_environment(config),
        working_directory=working_directory,
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )



import argparse
import collections
import enum
import json
import logging

import binaryornot.helpers

import borgmatic.borg.extract
import borgmatic.borg.list
import borgmatic.borg.repo_list
import borgmatic.borg.version

logger = logging.getLogger(__name__)


class Path_type(enum.Enum):
    DIRECTORY = 'd'
    LINK = 'l'
    PIPE = 'p'
    FILE = '-'


# A data structure capturing a path stored in a Borg archive.
Archive_path = collections.namedtuple(
    'Archive_path',
    ('path_type', 'file_path', 'link_target'),
)


def get_repository_archives(config, repository):
    '''
    Given a configuration dict and a repository dict, return a list of the repository's archives,
    one dict per archive.
    '''
    with borgmatic.logger.Log_prefix(repository.get('label', repository['path'])):
        logger.info('Listing repository')
        repo_list_arguments = argparse.Namespace(
            repository=repository['path'],
            short=None,
            format=None,
            json=True,
            prefix=None,
            match_archives=None,
            sort_by=None,
            first=None,
            last=None,
        )
        global_arguments = argparse.Namespace()
        local_path = config.get('local_path', 'borg')
        remote_path = config.get('remote_path')
        local_borg_version = borgmatic.borg.version.local_borg_version(config, local_path)

        return json.loads(
            borgmatic.borg.repo_list.list_repository(
                repository['path'],
                config,
                local_borg_version,
                repo_list_arguments,
                global_arguments,
                local_path,
                remote_path,
            )
        )


def get_archive_paths(config, repository, archive_name):
    '''
    Given a configuration dict, a repository dict, and an archive name in that repository, get a
    generator of files, directories, symlinks, etc. found in the archive, each as an Archive_path
    instance.
    '''
    with borgmatic.logger.Log_prefix(repository.get('label', repository['path'])):
        logger.info(f'Listing archive {archive_name}')

        global_arguments = argparse.Namespace()
        local_path = config.get('local_path', 'borg')
        remote_path = config.get('remote_path')
        local_borg_version = borgmatic.borg.version.local_borg_version(config, local_path)

        return (
            Archive_path(
                path_data['type'],
                path_data['path'],
                path_data.get('linktarget'),
            )
            for path_data in borgmatic.borg.list.capture_archive_listing(
                repository['path'],
                archive_name,
                config,
                local_borg_version,
                global_arguments,
                local_path=local_path,
                remote_path=remote_path,
            )
        )


READLINES_HINT_BYTES = 100000
TRUNCATION_MESSAGE = '[... truncated for display ...]'


def get_archive_file_content(config, repository, archive_name, file_path):
    '''
    Given a configuration dict, a repository dict, an archive name in that repository, and a file
    path in that archive, return the file's contents or None if the file can't be loaded, e.g.
    because it's binary or can't be decoded.

    If the file is too large, then truncate the returned content.
    '''
    with borgmatic.logger.Log_prefix(repository.get('label', repository['path'])):
        logger.info(f'Getting archive content of file {file_path}')
        local_path = config.get('local_path', 'borg')
        remote_path = config.get('remote_path')

        lines = borgmatic.borg.extract.extract_archive(
            dry_run=False,
            repository=repository['path'],
            archive=archive_name,
            paths=(file_path,),
            config=config,
            local_borg_version=borgmatic.borg.version.local_borg_version(config, local_path),
            global_arguments=argparse.Namespace(),
            local_path=local_path,
            remote_path=remote_path,
            destination_path=None,
            strip_components=None,
            extract_to_stdout=True,
        ).stdout.readlines(READLINES_HINT_BYTES)

        content = b''.join(lines)

        if binaryornot.helpers.is_binary_string(content):
            return None

        try:
            return (
                content.decode()
                if len(content) < READLINES_HINT_BYTES
                else f'{content.decode()}\n{TRUNCATION_MESSAGE}'
            )
        except UnicodeDecodeError:
            return None



import contextlib
import logging

import textual._context
import textual.widgets
import textual.worker

import borgmatic.logger


class Rich_color_formatter(logging.Formatter):
    '''
    A Python logging formatter that formats log records with Rich-compatible color markup according
    to their levels.
    '''

    def __init__(self, *args, **kwargs):
        self.prefix = None
        super().__init__(
            '{prefix}{message}',
            *args,
            style='{',
            **kwargs,
        )

    def format(self, record):
        '''
        Given a log record, format it with Rich-compatibe color markup corresponding to its log
        level.
        '''
        borgmatic.logger.add_custom_log_levels()

        color = {
            logging.CRITICAL: 'bright_red',
            logging.ERROR: 'bright_red',
            logging.WARNING: 'bright_yellow',
            logging.ANSWER: 'bright_magenta',
            logging.INFO: 'bright_green',
            logging.DEBUG: 'bright_cyan',
        }.get(record.levelno)
        record.prefix = f'{self.prefix}: ' if self.prefix else ''

        return f'[{color}]{super().format(record)}[/{color}]'


class Browse_log_handler(logging.Handler):
    '''
    A Python log handler that writes any log records to a logging widget.
    '''

    def __init__(self, logs_widget):
        '''
        Given a logs widget, save it for use below.
        '''
        self.logs_widget = logs_widget

        super().__init__()

    def emit(self, record):
        '''
        Given a log record, format it and log it to the logs widgets. This works whether or not the
        logging is happening in the main thread.
        '''
        message = self.format(record)

        try:
            textual.worker.get_current_worker()
            self.logs_widget.app.call_from_thread(self.logs_widget.write, message)
        except (RuntimeError, textual.worker.NoActiveWorker):
            with contextlib.suppress(textual._context.NoActiveAppError):
                self.logs_widget.write(message)


def log_to_widget(logs_widget):
    '''
    Given a Textual RichLog logs widget, add a log handler and formatter that logs to it. Also
    remove the default borgmatic console log handler so it doesn't try to log all over our UI.
    '''
    handler = Browse_log_handler(logs_widget)
    handler.setFormatter(Rich_color_formatter())
    logger = logging.getLogger()
    logger.setLevel(min(handler.level for handler in logger.handlers))
    logger.addHandler(handler)

    with contextlib.suppress(StopIteration):
        console_handler = next(
            handler
            for handler in logging.getLogger().handlers
            if isinstance(handler, borgmatic.logger.Multi_stream_handler)
        )
        logger.removeHandler(console_handler)


class Logs(textual.widgets.RichLog):
    '''
    A widget for viewing borgmatic logs in realtime. The log level is determined by borgmatic's
    current verbosity level.
    '''

    def __init__(self):
        super().__init__(markup=True, id='logs', classes='panel')
        self.border_title = '🪵 logs'



import logging

import requests

import borgmatic.hooks.credential.parse

logger = logging.getLogger(__name__)


EMERGENCY_PRIORITY = 2
TIMEOUT_SECONDS = 10


def initialize_monitor(
    ping_url,
    config,
    config_filename,
    monitoring_log_level,
    dry_run,
):  # pragma: no cover
    '''
    No initialization is necessary for this monitor.
    '''


def ping_monitor(hook_config, config, config_filename, state, monitoring_log_level, dry_run):
    '''
    Post a message to the configured Pushover application.
    If this is a dry run, then don't actually update anything.
    '''

    run_states = hook_config.get('states', ['fail'])

    if state.name.lower() not in run_states:
        return

    dry_run_label = ' (dry run; not actually updating)' if dry_run else ''

    state_config = hook_config.get(state.name.lower(), {})

    try:
        token = borgmatic.hooks.credential.parse.resolve_credential(
            hook_config.get('token'),
            config,
        )
        user = borgmatic.hooks.credential.parse.resolve_credential(hook_config.get('user'), config)
    except ValueError as error:
        logger.warning(f'Pushover credential error: {error}')
        return

    logger.info(f'Updating Pushover{dry_run_label}')

    if state_config.get('priority') == EMERGENCY_PRIORITY:
        if 'expire' not in state_config:
            logger.info('Setting expire to default (10 min)')
            state_config['expire'] = 600

        if 'retry' not in state_config:
            logger.info('Setting retry to default (30 sec)')
            state_config['retry'] = 30
    elif 'expire' in state_config or 'retry' in state_config:
        raise ValueError(
            'The configuration parameters retry and expire should not be set when priority is not equal to 2. Please remove them from the configuration.',
        )

    state_config = {
        key: (int(value) if key == 'html' else value) for key, value in state_config.items()
    }

    data = dict(
        {
            'token': token,
            'user': user,
            # Default to state name. Can be overwritten by state_config below.
            'message': state.name.lower(),
        },
        **state_config,
    )

    if not dry_run:
        logging.getLogger('urllib3').setLevel(logging.ERROR)
        try:
            response = requests.post(
                'https://api.pushover.net/1/messages.json',
                data=data,
                timeout=TIMEOUT_SECONDS,
                headers={
                    'Content-type': 'application/x-www-form-urlencoded',
                    'User-Agent': 'borgmatic',
                },
            )
            if not response.ok:
                response.raise_for_status()
        except requests.exceptions.RequestException as error:
            logger.warning(f'Pushover error: {error}')


def destroy_monitor(ping_url_or_uuid, config, monitoring_log_level, dry_run):  # pragma: no cover
    '''
    No destruction is necessary for this monitor.
    '''



import collections
import functools
import glob
import logging
import os
import pathlib
import shutil
import subprocess

import borgmatic.borg.pattern
import borgmatic.config.paths
import borgmatic.execute
import borgmatic.hooks.data_source.config
import borgmatic.hooks.data_source.snapshot

logger = logging.getLogger(__name__)


def use_streaming(hook_config, config):  # pragma: no cover
    '''
    Return whether dump streaming is used for this hook. (Spoiler: It isn't.)
    '''
    return False


BTRFS_SUBVOLUME_INODE_NUMBER = 256


@functools.cache
def path_is_a_subvolume(path):
    '''
    Given a path, return whether it is a Btrfs subvolume. Return False if the path doesn't exist.

    As a performance optimization, multiple calls to this function with the same path are cached.
    '''
    if path == os.environ.get('BTRFS_TEST_SUBVOLUME_PATH'):  # pragma: no cover
        return True

    try:
        return os.stat(path).st_ino == BTRFS_SUBVOLUME_INODE_NUMBER
    except FileNotFoundError:
        return False


@functools.cache
def get_subvolume_property(btrfs_command, subvolume_path, property_name):
    '''
    Given a btrfs command, a subvolume path, and a property name to lookup, return the value of the
    corresponding property.

    Raise subprocess.CalledProcessError if the btrfs command errors.

    As a performance optimization, multiple calls to this function with the same arguments are
    cached.
    '''
    output = '\n'.join(
        borgmatic.execute.execute_command_and_capture_output(
            (
                *btrfs_command.split(' '),
                'property',
                'get',
                '-t',  # Type.
                'subvol',
                subvolume_path,
                property_name,
            ),
            close_fds=True,
        )
    )

    try:
        value = output.strip().split('=')[1]
    except IndexError:
        raise ValueError(f'Invalid {btrfs_command} property output')

    return {
        'true': True,
        'false': False,
    }.get(value, value)


def get_containing_subvolume_path(btrfs_command, path):
    '''
    Given a btrfs command and a path, return the subvolume path that contains the given path (or is
    the same as the path).

    If there is no such subvolume path or the containing subvolume is read-only, return None.
    '''
    # Probe the given pattern's path and all of its parents, grandparents, etc. to try to find a
    # Btrfs subvolume.
    for candidate_path in (
        path,
        *tuple(str(ancestor) for ancestor in pathlib.PurePath(path).parents),
    ):
        if not path_is_a_subvolume(candidate_path):
            continue

        try:
            if get_subvolume_property(btrfs_command, candidate_path, 'ro'):
                logger.debug(f'Ignoring Btrfs subvolume {candidate_path} because it is read-only')

                return None

            logger.debug(f'Path {candidate_path} is a Btrfs subvolume')

            return candidate_path
        except subprocess.CalledProcessError as error:
            logger.debug(
                f'Error determining read-only status of Btrfs subvolume {candidate_path}: {error}',
            )

            return None

    return None


def get_all_subvolume_paths(btrfs_command, patterns):
    '''
    Given a btrfs command and a sequence of patterns, get the sorted paths for all Btrfs subvolumes
    containing those patterns.
    '''
    return tuple(
        sorted(
            {
                subvolume_path
                for pattern in patterns
                if pattern.type == borgmatic.borg.pattern.Pattern_type.ROOT
                if pattern.source == borgmatic.borg.pattern.Pattern_source.CONFIG
                for subvolume_path in (get_containing_subvolume_path(btrfs_command, pattern.path),)
                if subvolume_path
            }
        ),
    )


Subvolume = collections.namedtuple('Subvolume', ('path', 'contained_patterns'), defaults=((),))


def get_subvolumes(btrfs_command, patterns):
    '''
    Given a Btrfs command to run and a sequence of configured patterns, find the intersection
    between the current Btrfs filesystem/subvolume paths and the paths of any patterns. The idea is
    that these pattern paths represent the requested subvolumes to snapshot.

    Only include subvolumes that contain at least one root pattern sourced from borgmatic
    configuration (as opposed to generated elsewhere in borgmatic).

    Return the result as a sequence of matching Subvolume instances.
    '''
    candidate_patterns = set(patterns or ())
    subvolumes = []

    # For each subvolume path, match it against the given patterns to find the subvolumes to
    # backup. Sort the subvolumes from longest to shortest mount points, so longer subvolumes get
    # a whack at the candidate pattern piñata before their parents do. (Patterns are consumed during
    # this process, so no two subvolumes end up with the same contained patterns.)
    for subvolume_path in reversed(get_all_subvolume_paths(btrfs_command, patterns)):
        subvolumes.extend(
            Subvolume(subvolume_path, contained_patterns)
            for contained_patterns in (
                borgmatic.hooks.data_source.snapshot.get_contained_patterns(
                    subvolume_path,
                    candidate_patterns,
                ),
            )
            if any(
                pattern.type == borgmatic.borg.pattern.Pattern_type.ROOT
                and pattern.source == borgmatic.borg.pattern.Pattern_source.CONFIG
                for pattern in contained_patterns
            )
        )

    return tuple(sorted(subvolumes, key=lambda subvolume: subvolume.path))


BORGMATIC_SNAPSHOT_PREFIX = '.borgmatic-snapshot'


def make_snapshot_path(subvolume_path):
    '''
    Given the path to a subvolume, make a corresponding snapshot path for it.
    '''
    return os.path.join(
        subvolume_path,
        f'{BORGMATIC_SNAPSHOT_PREFIX}',
        # Included so that the snapshot ends up in the Borg archive at the "original" subvolume path.
    ) + subvolume_path.rstrip(os.path.sep)


def make_snapshot_exclude_pattern(subvolume_path):  # pragma: no cover
    '''
    Given the path to a subvolume, make a corresponding exclude pattern for its embedded snapshot
    path. This is to work around a quirk of Btrfs: If you make a snapshot path as a child directory
    of a subvolume, then the snapshot's own initial directory component shows up as an empty
    directory within the snapshot itself. For instance, if you have a Btrfs subvolume at /mnt and
    make a snapshot of it at:

        /mnt/.borgmatic-snapshot/mnt

    ... then the snapshot itself will have an empty directory at:

        /mnt/.borgmatic-snapshot/mnt/.borgmatic-snapshot

    So to prevent that from ending up in the Borg archive, this function produces an exclude pattern
    to exclude that path.
    '''
    snapshot_directory = f'{BORGMATIC_SNAPSHOT_PREFIX}'

    return borgmatic.borg.pattern.Pattern(
        os.path.join(
            subvolume_path,
            snapshot_directory,
            subvolume_path.lstrip(os.path.sep),
            snapshot_directory,
        ),
        borgmatic.borg.pattern.Pattern_type.NO_RECURSE,
        borgmatic.borg.pattern.Pattern_style.FNMATCH,
        source=borgmatic.borg.pattern.Pattern_source.HOOK,
    )


def make_borg_snapshot_pattern(subvolume_path, pattern):
    '''
    Given the path to a subvolume and a pattern as a borgmatic.borg.pattern.Pattern instance whose
    path is inside the subvolume, return a new Pattern with its path rewritten to be in a snapshot
    path intended for giving to Borg.

    Move any initial caret in a regular expression pattern path to the beginning, so as not to break
    the regular expression.
    '''
    initial_caret = (
        '^'
        if pattern.style == borgmatic.borg.pattern.Pattern_style.REGULAR_EXPRESSION
        and pattern.path.startswith('^')
        else ''
    )

    rewritten_path = initial_caret + os.path.join(
        subvolume_path,
        f'{BORGMATIC_SNAPSHOT_PREFIX}',
        # Use the Borg 1.4+ "slashdot" hack to prevent the snapshot path prefix from getting
        # included in the archive—but only if there's not already a slashdot hack present in the
        # pattern.
        ('' if f'{os.path.sep}.{os.path.sep}' in pattern.path else '.'),
        # Included so that the source directory ends up in the Borg archive at its "original" path.
        pattern.path.lstrip('^').lstrip(os.path.sep),
    )

    return borgmatic.borg.pattern.Pattern(
        rewritten_path,
        pattern.type,
        pattern.style,
        pattern.device,
        source=borgmatic.borg.pattern.Pattern_source.HOOK,
    )


def snapshot_subvolume(btrfs_command, subvolume_path, snapshot_path):  # pragma: no cover
    '''
    Given a Btrfs command to run, the path to a subvolume, and the path for a snapshot, create a new
    Btrfs snapshot of the subvolume.
    '''
    os.makedirs(os.path.dirname(snapshot_path), mode=0o700, exist_ok=True)

    borgmatic.execute.execute_command(
        (
            *btrfs_command.split(' '),
            'subvolume',
            'snapshot',
            subvolume_path,
            snapshot_path,
        ),
        output_log_level=logging.DEBUG,
        close_fds=True,
    )


def dump_data_sources(
    hook_config,
    config,
    config_paths,
    borgmatic_runtime_directory,
    patterns,
    dry_run,
):
    '''
    Given a Btrfs configuration dict, a configuration dict, the borgmatic configuration file paths,
    the borgmatic runtime directory, the configured patterns, and whether this is a dry run,
    auto-detect and snapshot any Btrfs subvolume paths listed in the given patterns. Also update
    those patterns, replacing subvolume paths with corresponding snapshot directories so they get
    stored in the Borg archive instead.

    Return an empty sequence, since there are no ongoing dump processes from this hook.

    If this is a dry run, then don't actually snapshot anything.
    '''
    dry_run_label = ' (dry run; not actually snapshotting anything)' if dry_run else ''
    logger.info(f'Snapshotting Btrfs subvolumes{dry_run_label}')

    if 'findmnt_command' in hook_config:
        logger.warning(
            'The Btrfs "findmnt_command" option is deprecated and will be removed from a future release; findmnt is no longer used',
        )

    # Based on the configured patterns, determine Btrfs subvolumes to backup. Only consider those
    # patterns that came from actual user configuration (as opposed to, say, other hooks).
    btrfs_command = hook_config.get('btrfs_command', 'btrfs')
    subvolumes = get_subvolumes(btrfs_command, patterns)

    if not subvolumes:
        logger.warning(f'No Btrfs subvolumes found to snapshot{dry_run_label}')

    # Snapshot each subvolume, rewriting patterns to use their snapshot paths.
    for subvolume in subvolumes:
        logger.debug(f'Creating Btrfs snapshot for {subvolume.path} subvolume')

        snapshot_path = make_snapshot_path(subvolume.path)

        if dry_run:
            continue

        snapshot_subvolume(btrfs_command, subvolume.path, snapshot_path)

        for pattern in subvolume.contained_patterns:
            snapshot_pattern = make_borg_snapshot_pattern(subvolume.path, pattern)
            borgmatic.hooks.data_source.config.replace_pattern(patterns, pattern, snapshot_pattern)

        borgmatic.hooks.data_source.config.inject_pattern(
            patterns, make_snapshot_exclude_pattern(subvolume.path)
        )

    return []


def delete_snapshot(btrfs_command, snapshot_path):  # pragma: no cover
    '''
    Given a Btrfs command to run and the name of a snapshot path, delete it.
    '''
    borgmatic.execute.execute_command(
        (
            *btrfs_command.split(' '),
            'subvolume',
            'delete',
            snapshot_path,
        ),
        output_log_level=logging.DEBUG,
        close_fds=True,
    )


def remove_data_source_dumps(hook_config, config, borgmatic_runtime_directory, patterns, dry_run):
    '''
    Given a Btrfs configuration dict, a configuration dict, the borgmatic runtime directory, the
    configured patterns, and whether this is a dry run, delete any Btrfs snapshots created by
    borgmatic. If this is a dry run or Btrfs isn't configured in borgmatic's configuration, then
    don't actually remove anything.
    '''
    if hook_config is None:
        return

    dry_run_label = ' (dry run; not actually removing anything)' if dry_run else ''

    btrfs_command = hook_config.get('btrfs_command', 'btrfs')

    try:
        all_subvolumes = get_subvolumes(btrfs_command, patterns)
    except FileNotFoundError as error:
        logger.debug(f'Could not find "{error.filename}" command')
        return
    except subprocess.CalledProcessError as error:
        logger.debug(error)
        return

    # Reversing the sorted subvolumes ensures that we remove longer paths of child subvolumes before
    # the shorter paths of parent subvolumes.
    for subvolume in reversed(all_subvolumes):
        subvolume_snapshots_glob = borgmatic.config.paths.replace_temporary_subdirectory_with_glob(
            os.path.normpath(make_snapshot_path(subvolume.path)),
            temporary_directory_prefix=BORGMATIC_SNAPSHOT_PREFIX,
        )

        logger.debug(
            f'Looking for snapshots to remove in {subvolume_snapshots_glob}{dry_run_label}',
        )

        for snapshot_path in glob.glob(subvolume_snapshots_glob):
            if not os.path.isdir(snapshot_path):
                continue

            logger.debug(f'Deleting Btrfs snapshot {snapshot_path}{dry_run_label}')

            if dry_run:
                continue

            try:
                delete_snapshot(btrfs_command, snapshot_path)
            except FileNotFoundError:
                logger.debug(f'Could not find "{btrfs_command}" command')
                return
            except subprocess.CalledProcessError as error:
                logger.debug(error)
                return

            # Remove the snapshot parent directory if it still exists. (It might not exist if the
            # snapshot was for "/".)
            snapshot_parent_dir = snapshot_path.rsplit(subvolume.path, 1)[0]

            if os.path.isdir(snapshot_parent_dir):
                shutil.rmtree(snapshot_parent_dir)


def make_data_source_dump_patterns(
    hook_config,
    config,
    borgmatic_runtime_directory,
    name=None,
    hostname=None,
    port=None,
    container=None,
    label=None,
):  # pragma: no cover
    '''
    Restores aren't implemented, because stored files can be extracted directly with "extract".
    '''
    return ()


def restore_data_source_dump(
    hook_config,
    config,
    data_source,
    dry_run,
    extract_process,
    connection_params,
    borgmatic_runtime_directory,
):  # pragma: no cover
    '''
    Restores aren't implemented, because stored files can be extracted directly with "extract".
    '''
    raise NotImplementedError()



import logging

import borgmatic.borg.prune

logger = logging.getLogger(__name__)


def run_prune(
    config_filename,
    repository,
    config,
    local_borg_version,
    prune_arguments,
    global_arguments,
    dry_run_label,
    local_path,
    remote_path,
):
    '''
    Run the "prune" action for the given repository.
    '''
    logger.info(f'Pruning archives{dry_run_label}')
    borgmatic.borg.prune.prune_archives(
        global_arguments.dry_run,
        repository['path'],
        config,
        local_borg_version,
        prune_arguments,
        global_arguments,
        local_path=local_path,
        remote_path=remote_path,
    )



import logging


def run_browse(
    diff_arguments,
    global_arguments,
    configs,
):
    '''
    Run the "browse" action for the given borgmatic configurations. This launches a console UI.

    Raise ValueError if the Textual library (a prerequisite for this action) can't be imported.
    '''
    if not configs:
        return

    logging.getLogger('asyncio').setLevel(logging.WARNING)

    try:
        import textual  # noqa: F401, PLC0415
    except ImportError:  # pragma: no cover
        raise ValueError(
            'Unable to import the Textual library for the browse action; try installing "borgmatic[browse]"'
        )

    import borgmatic.actions.browse.app  # noqa: PLC0415

    app = borgmatic.actions.browse.app.Browse_app(configs)
    app.run()



import textual.binding
import textual.widgets

OPTION_LIST_BINDINGS = (
    *textual.widgets.OptionList.BINDINGS,
    textual.binding.Binding(
        key='up,k', action='cursor_up', description='scroll up', show=True, priority=True
    ),
    textual.binding.Binding(
        key='down,j', action='cursor_down', description='scroll down', show=True, priority=True
    ),
    textual.binding.Binding(
        key='pageup', action='page_up', description='page up', show=True, priority=True
    ),
    textual.binding.Binding(
        key='pagedown', action='page_down', description='page down', show=True, priority=True
    ),
    textual.binding.Binding(
        key='enter', action='select', description='select', show=True, priority=True
    ),
    textual.binding.Binding(key='right,l', action='select', description='select', show=False),
)



import logging

import borgmatic.borg.change_passphrase

logger = logging.getLogger(__name__)


def run_change_passphrase(
    repository,
    config,
    local_borg_version,
    change_passphrase_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "key change-passphrase" action for the given repository.
    '''
    logger.info('Changing repository passphrase')
    borgmatic.borg.change_passphrase.change_passphrase(
        repository['path'],
        config,
        local_borg_version,
        change_passphrase_arguments,
        global_arguments,
        local_path=local_path,
        remote_path=remote_path,
    )



import logging
import shlex

import borgmatic.commands.arguments
import borgmatic.config.paths
import borgmatic.logger
from borgmatic.borg import environment, flags
from borgmatic.execute import DO_NOT_CAPTURE, execute_command

logger = logging.getLogger(__name__)


BORG_SUBCOMMANDS_WITH_SUBCOMMANDS = {'key', 'debug'}


def run_arbitrary_borg(
    repository_path,
    config,
    local_borg_version,
    options,
    archive=None,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a local or remote repository path, a configuration dict, the local Borg version, a
    sequence of arbitrary command-line Borg options, and an optional archive name, run an arbitrary
    Borg command, passing in REPOSITORY and ARCHIVE environment variables for optional use in the
    command.
    '''
    borgmatic.logger.add_custom_log_levels()

    try:
        options = options[1:] if options[0] == '--' else options

        # Borg commands like "key" have a sub-command ("export", etc.) that must follow it.
        command_options_start_index = 2 if options[0] in BORG_SUBCOMMANDS_WITH_SUBCOMMANDS else 1
        borg_command = tuple(options[:command_options_start_index])
        command_options = tuple(options[command_options_start_index:])

        if borg_command and borg_command[0] in borgmatic.commands.arguments.ACTION_ALIASES:
            logger.warning(
                f"Borg's {borg_command[0]} subcommand is supported natively by borgmatic. Try this instead: borgmatic {borg_command[0]}",
            )
    except IndexError:
        borg_command = ()
        command_options = ()

    full_command = (
        (local_path,)
        + borg_command
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug', '--show-rc') if logger.isEnabledFor(logging.DEBUG) else ())
        + flags.make_flags('remote-path', remote_path)
        + flags.make_flags('lock-wait', config.get('lock_wait'))
        + command_options
    )

    return execute_command(  # noqa: S604
        tuple(shlex.quote(part) for part in full_command),
        output_file=DO_NOT_CAPTURE,
        shell=True,
        environment=dict(
            (environment.make_environment(config) or {}),
            BORG_REPO=repository_path,
            ARCHIVE=archive if archive else '',
        ),
        working_directory=borgmatic.config.paths.get_working_directory(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )



import textual.widgets

import borgmatic.actions.browse.bindings
import borgmatic.actions.browse.loading
import borgmatic.actions.browse.workers


class Archives_list(textual.widgets.OptionList):
    '''
    A widget for selecting a single Borg archive from among the archives in a repository. The item
    selection event is handled in a Carousel instance, the parent widget of an Archives_list.
    '''

    BINDINGS = borgmatic.actions.browse.bindings.OPTION_LIST_BINDINGS

    def __init__(self, config, repository):
        '''
        Given a configuration dict and a repository dict, prepare to load the archives from the
        repository for eventual display in this widget. Actual loading kicks off in on_mount()
        below.
        '''
        self.config = config
        self.repository = repository

        super().__init__(classes='panel')
        self.border_title = '📚 archives'
        self.highlighted_option_changed = False
        self.archive_loaded = borgmatic.actions.browse.workers.Archive_loaded(
            self, 'archive loaded'
        )

        self.loading_timer = borgmatic.actions.browse.loading.add_inline_loading_indicator(self)

    def on_mount(self):
        '''
        When this widget gets mounted in the DOM, subscribe to archive loaded events so that we can
        find out about archives as they load. Also start loading archives from the repository.

        Loading is started *after* subscribing to the archive loaded signal so that there's not a
        gap where we might miss out on signal publishes.
        '''
        self.archive_loaded.subscribe(self, self.on_archive_loaded)

        borgmatic.actions.browse.workers.add_repository_archives(
            self.app,
            archive_loaded=self.archive_loaded,
            config=self.config,
            repository=self.repository,
        )

    def on_archive_loaded(self, archive_name):
        '''
        When an archive loads, add it as an option to this archives list. But if we get a
        signal that all path loading is complete, stop and remove our loading indicator.
        '''
        if archive_name is borgmatic.actions.browse.workers.LOADING_DONE:
            self.loading_timer.stop()
            self.remove_option('loading-indicator')
            return

        label_pieces = (
            (archive_name, '[dim](latest)[/dim]') if len(self.options) == 1 else (archive_name,)
        )
        highlighted_option = self.highlighted_option

        loading_indicator = self.get_option('loading-indicator')
        self.remove_option('loading-indicator')
        self.add_options(
            (
                textual.widgets.option_list.Option(' '.join(label_pieces), id=archive_name),
                loading_indicator,
            ),
        )

        # Retain the highlighted option position even as other options load around it.
        self.highlighted = (
            self.get_option_index(highlighted_option.id)
            if highlighted_option and self.highlighted_option_changed
            else 0
        )

    def on_option_list_option_highlighted(self, event):
        '''
        When the highlighted option changes, record that fact. This flag is consumed in
        borgmatic.actions.browse.workers.add_repository_archives() in order to retain the
        highlighted option even as other options load around it.
        '''
        if self.highlighted not in {None, 0}:
            self.highlighted_option_changed = True



import borgmatic.commands.arguments


def upgrade_message(language: str, upgrade_command: str, completion_file: str):
    return f'''
Your {language} completions script is from a different version of borgmatic than is
currently installed. Please upgrade your script so your completions match the
command-line flags in your installed borgmatic! Try this to upgrade:

    {upgrade_command}
    source {completion_file}
'''


def available_actions(subparsers, current_action=None):
    '''
    Given subparsers as an argparse._SubParsersAction instance and a current action name (if
    any), return the actions names that can follow the current action on a command-line.

    This takes into account which sub-actions that the current action supports. For instance, if
    "bootstrap" is a sub-action for "config", then "bootstrap" should be able to follow a current
    action of "config" but not "list".
    '''
    action_to_subactions = borgmatic.commands.arguments.get_subactions_for_actions(
        subparsers.choices,
    )
    current_subactions = action_to_subactions.get(current_action)

    if current_subactions:
        return current_subactions

    all_subactions = {
        subaction for subactions in action_to_subactions.values() for subaction in subactions
    }

    return tuple(action for action in subparsers.choices if action not in all_subactions)



import itertools
import logging
import re

from borgmatic.borg import feature

logger = logging.getLogger(__name__)


def make_flags(name, value):
    '''
    Given a flag name and its value, return it formatted as Borg-compatible flags.
    '''
    if not value:
        return ()

    flag = f"--{name.replace('_', '-')}"

    if value is True:
        return (flag,)

    return (flag, str(value))


def make_flags_from_arguments(arguments, excludes=()):
    '''
    Given borgmatic command-line arguments as an instance of argparse.Namespace, and optionally a
    list of named arguments to exclude, generate and return the corresponding Borg command-line
    flags as a tuple.
    '''
    return tuple(
        itertools.chain.from_iterable(
            make_flags(name, value=getattr(arguments, name))
            for name in sorted(vars(arguments))
            if name not in excludes and not name.startswith('_')
        ),
    )


def make_repository_flags(repository_path, local_borg_version):
    '''
    Given the path of a Borg repository and the local Borg version, return Borg-version-appropriate
    command-line flags (as a tuple) for selecting that repository.
    '''
    return (
        ('--repo',)
        if feature.available(feature.Feature.SEPARATE_REPOSITORY_ARCHIVE, local_borg_version)
        else ()
    ) + (repository_path,)


ARCHIVE_HASH_PATTERN = re.compile(r'[0-9a-fA-F]{8,}$')


def make_repository_archive_flags(repository_path, archive, local_borg_version):
    '''
    Given the path of a Borg repository, an archive name or pattern, and the local Borg version,
    return Borg-version-appropriate command-line flags (as a tuple) for selecting that repository
    and archive.
    '''
    return (
        (
            '--repo',
            repository_path,
            (
                f'aid:{archive}'
                if feature.available(feature.Feature.ARCHIVE_SERIES, local_borg_version)
                and ARCHIVE_HASH_PATTERN.match(archive)
                and not archive.startswith('aid:')
                else archive
            ),
        )
        if feature.available(feature.Feature.SEPARATE_REPOSITORY_ARCHIVE, local_borg_version)
        else (f'{repository_path}::{archive}',)
    )


DEFAULT_ARCHIVE_NAME_FORMAT_WITHOUT_SERIES = '{hostname}-{now:%Y-%m-%dT%H:%M:%S.%f}'
DEFAULT_ARCHIVE_NAME_FORMAT_WITH_SERIES = '{hostname}'


def get_default_archive_name_format(local_borg_version):
    '''
    Given the local Borg version, return the corresponding default archive name format.
    '''
    if feature.available(feature.Feature.ARCHIVE_SERIES, local_borg_version):
        return DEFAULT_ARCHIVE_NAME_FORMAT_WITH_SERIES

    return DEFAULT_ARCHIVE_NAME_FORMAT_WITHOUT_SERIES


def make_match_archives_flags(  # noqa: PLR0911
    match_archives,
    archive_name_format,
    local_borg_version,
    default_archive_name_format=None,
    force_flags_even_for_globs=False,
):
    '''
    Return match archives flags based on the given match archives value, if any. If it isn't set,
    return match archives flags to match archives created with the given (or default) archive name
    format. This is done by replacing certain archive name format placeholders for ephemeral data
    (like "{now}") with globs.

    If the given match archives value is just a glob (e.g. "*"), omit any flags—unless the force
    flags even for globs value is True.
    '''
    if match_archives:
        if not force_flags_even_for_globs and match_archives in {'*', 're:.*', 'sh:*'}:
            return ()

        if feature.available(feature.Feature.MATCH_ARCHIVES, local_borg_version):
            if (
                feature.available(feature.Feature.ARCHIVE_SERIES, local_borg_version)
                and ARCHIVE_HASH_PATTERN.match(match_archives)
                and not match_archives.startswith('aid:')
            ):
                return ('--match-archives', f'aid:{match_archives}')

            return ('--match-archives', match_archives)

        return ('--glob-archives', re.sub(r'^sh:', '', match_archives))

    derived_match_archives = re.sub(
        r'\{(now|utcnow|pid)([:%\w\.-]*)\}',
        '*',
        archive_name_format
        or default_archive_name_format
        or get_default_archive_name_format(local_borg_version),
    )

    if not force_flags_even_for_globs and derived_match_archives == '*':
        return ()

    if feature.available(feature.Feature.MATCH_ARCHIVES, local_borg_version):
        return ('--match-archives', f'sh:{derived_match_archives}')

    return ('--glob-archives', f'{derived_match_archives}')


def warn_for_aggressive_archive_flags(command, output_lines):
    '''
    Given an archives command and the resulting output lines from running it, warn if the command
    used an archive flag but the output indicates zero archives were found.
    '''
    if {'--glob-archives', '--match-archives'}.intersection(set(command)) and len(
        tuple(line for line in output_lines if not line.startswith('terminating with '))
    ) == 0:
        logger.warning('An archive filter was applied, but no matching archives were found.')
        logger.warning(
            'Try adding --match-archives "*" or adjusting archive_name_format/match_archives in configuration.',
        )


def omit_flag(arguments, flag):
    '''
    Given a sequence of Borg command-line arguments, return them with the given (valueless) flag
    omitted. For instance, if the flag is "--flag" and arguments is:

        ('borg', 'create', '--flag', '--other-flag')

    ... then return:

        ('borg', 'create', '--other-flag')
    '''
    return tuple(argument for argument in arguments if argument != flag)


def omit_flag_and_value(arguments, flag):
    '''
    Given a sequence of Borg command-line arguments, return them with the given flag and its
    corresponding value omitted. For instance, if the flag is "--flag" and arguments is:

        ('borg', 'create', '--flag', 'value', '--other-flag')

    ... or:

        ('borg', 'create', '--flag=value', '--other-flag')

    ... then return:

        ('borg', 'create', '--other-flag')
    '''
    # This works by zipping together a list of overlapping pairwise arguments. E.g., ('one', 'two',
    # 'three', 'four') becomes ((None, 'one'), ('one, 'two'), ('two', 'three'), ('three', 'four')).
    # This makes it easy to "look back" at the previous arguments so we can exclude both a flag and
    # its value.
    return tuple(
        argument
        for (previous_argument, argument) in zip((None, *arguments), arguments)
        if flag not in {previous_argument, argument}
        if not argument.startswith(f'{flag}=')
    )


def make_exclude_flags(config):
    '''
    Given a configuration dict with various exclude options, return the corresponding Borg flags as
    a tuple.
    '''
    caches_flag = ('--exclude-caches',) if config.get('exclude_caches') else ()
    if_present_flags = tuple(
        itertools.chain.from_iterable(
            ('--exclude-if-present', if_present)
            for if_present in config.get('exclude_if_present', ())
        ),
    )
    keep_exclude_tags_flags = ('--keep-exclude-tags',) if config.get('keep_exclude_tags') else ()
    exclude_nodump_flags = ('--exclude-nodump',) if config.get('exclude_nodump') else ()

    return caches_flag + if_present_flags + keep_exclude_tags_flags + exclude_nodump_flags


def make_list_filter_flags(local_borg_version, dry_run):
    '''
    Given the local Borg version and whether this is a dry run, return the corresponding flags for
    passing to "--list --filter". The general idea is that excludes are shown for a dry run or when
    the verbosity is debug.
    '''
    base_flags = 'AME'
    show_excludes = logger.isEnabledFor(logging.DEBUG)

    if feature.available(feature.Feature.EXCLUDED_FILES_MINUS, local_borg_version):
        if show_excludes or dry_run:
            return f'{base_flags}+-'

        return base_flags

    if show_excludes:
        return f'{base_flags}x-'

    return f'{base_flags}-'



import fnmatch
import json
import logging
import os
import re
import shutil

import borgmatic.actions.restore

logger = logging.getLogger(__name__)

IS_A_HOOK = False


def make_data_source_dump_path(borgmatic_runtime_directory, data_source_hook_name):
    '''
    Given a borgmatic runtime directory and a data source hook name, construct a data source dump
    path.
    '''
    return os.path.join(borgmatic_runtime_directory, data_source_hook_name)


def make_data_source_dump_filename(
    dump_path, name, hostname=None, port=None, container=None, label=None
):
    '''
    Based on the given dump directory path, data source name, hostname, and port, return a filename
    to use for the data source dump. The hostname defaults to localhost.

    Raise ValueError if the data source name is invalid.
    '''
    if os.path.sep in name:
        raise ValueError(f'Invalid data source name {name}')

    if container:
        hostname = container
    elif not hostname:
        hostname = 'localhost'

    identifier = label if label else hostname + ('' if port is None else f':{port}')

    return os.path.join(dump_path, identifier, name)


def write_data_source_dumps_metadata(borgmatic_runtime_directory, hook_name, dumps_metadata):
    '''
    Given the borgmatic runtime directory, a data source hook name, and a sequence of
    borgmatic.actions.restore.Dump instances of dump metadata, write a metadata file describing all
    of those dumps. This metadata is being dumped so that it's available upon restore, e.g. to
    support the user selecting which data source(s) should be restored.

    Raise ValueError if writing to the file results in an operating system error.
    '''
    dumps_metadata_path = os.path.join(borgmatic_runtime_directory, hook_name, 'dumps.json')

    try:
        with open(dumps_metadata_path, 'w', encoding='utf-8') as metadata_file:
            json.dump(
                {
                    'dumps': [dump._asdict() for dump in dumps_metadata],
                },
                metadata_file,
                sort_keys=True,
            )
    except OSError as error:
        raise ValueError(f'Error writing to dumps metadata at {dumps_metadata_path}: {error}')


def parse_data_source_dumps_metadata(dumps_json, dumps_metadata_path):
    '''
    Given a dumps metadata JSON string as extracted from an archive and its path within the archive,
    parse it into a tuple of borgmatic.actions.restore.Dump instances and return them.

    Raise ValueError if parsing the JSON results in a JSON decode error or the data does not have
    the expected keys.
    '''
    try:
        return tuple(
            borgmatic.actions.restore.Dump(**dump) for dump in json.loads(dumps_json)['dumps']
        )
    except (json.JSONDecodeError, TypeError, KeyError) as error:
        raise ValueError(
            f'Cannot read archive data source dumps metadata at {dumps_metadata_path} due to invalid JSON: {error}',
        )


def create_parent_directory_for_dump(dump_path):
    '''
    Create a directory to contain the given dump path.
    '''
    os.makedirs(os.path.dirname(dump_path), mode=0o700, exist_ok=True)


def create_named_pipe_for_dump(dump_path):
    '''
    Create a named pipe at the given dump path.
    '''
    create_parent_directory_for_dump(dump_path)
    os.mkfifo(dump_path, mode=0o600)


def remove_data_source_dumps(dump_path, data_source_type_name, dry_run):
    '''
    Remove all data source dumps in the given dump directory path (including the directory itself).
    If this is a dry run, then don't actually remove anything.
    '''
    dry_run_label = ' (dry run; not actually removing anything)' if dry_run else ''

    logger.debug(f'Removing {data_source_type_name} data source dumps{dry_run_label}')

    if dry_run:
        return

    if os.path.exists(dump_path):
        shutil.rmtree(dump_path)


END_OF_STRING_PATTTERN = re.compile(r'\\z', flags=re.IGNORECASE)


def convert_glob_patterns_to_borg_pattern(patterns):
    '''
    Convert a sequence of shell glob patterns like "/etc/*", "/tmp/*" to the corresponding Borg
    regular expression archive pattern as a single string like "re:etc/.*|tmp/.*".
    '''
    # This deserves some explanation. Here's what this code is doing for each shell glob pattern:
    #
    #  * Strip off the leading slash, since Borg doesn't store leading slashes in archives.
    #  * Use fnmatch.translate() to convert the pattern to a Borg pattern.
    #  * Remove the "\z" or \"Z" end-of-string special character generated by fnmatch.translate(),
    #    because Borg doesn't like it. Replace it with a "$" end-of-string special character
    #    instead. And yes, this is using a regular expression to modify a regular expression.
    #  * Do the above for each of:
    #    * The plain pattern. This supports the use case of a standard database where its name is a
    #      filename in the path. Example: borgmatic/foo_databases/hostname/test
    #    * The pattern with "/*" tacked onto the end of it. This supports the use case of a
    #      directory-format database where its name is a directory name in the path. Example:
    #      borgmatic/foo_databases/hostname/test/toc.dat
    #
    # Join the resulting transformed patterns together with "|" and return them as a string.
    return 're:' + '|'.join(
        re.sub(END_OF_STRING_PATTTERN, '$', fnmatch.translate(stripped))
        + '|'
        + re.sub(END_OF_STRING_PATTTERN, '$', fnmatch.translate(stripped + '/*'))
        for pattern in patterns
        for stripped in (pattern.lstrip('/'),)
    )



import logging
import os

logger = logging.getLogger(__name__)


def load_credential(hook_config, config, credential_parameters):
    '''
    Given the hook configuration dict, the configuration dict, and a credential parameters tuple
    containing a credential path to load, load the credential from file and return it.

    Raise ValueError if the credential parameters is not one element or the secret file cannot be
    read.
    '''
    try:
        (credential_path,) = credential_parameters
    except ValueError:
        name = ' '.join(credential_parameters)

        raise ValueError(f'Cannot load invalid credential: "{name}"')

    expanded_credential_path = os.path.expanduser(credential_path)

    try:
        with open(
            os.path.join(config.get('working_directory', ''), expanded_credential_path),
            encoding='utf-8',
        ) as credential_file:
            return credential_file.read().rstrip(os.linesep)
    except (FileNotFoundError, OSError) as error:
        logger.warning(error)

        raise ValueError(f'Cannot load credential file: {error.filename}')



import borgmatic.actions.browse.archive

PATH_TYPE_ICONS = {
    borgmatic.actions.browse.archive.Path_type.DIRECTORY.value: '📁',
    borgmatic.actions.browse.archive.Path_type.LINK.value: '🔗',
    borgmatic.actions.browse.archive.Path_type.PIPE.value: '🚰',
    borgmatic.actions.browse.archive.Path_type.FILE.value: '📄',
}



import contextlib
import functools
import logging

import textual.widgets
import textual.widgets.option_list

LOADING_DOT_INTERVAL_SECONDS = 0.3


logger = logging.getLogger('__name__')


def update_inline_loading_indicator(widget):
    '''
    Given a textual.widgets.OptionList or a textual.widgets.RichLog instance, animate the existing
    loading indicator inside it.
    '''
    if isinstance(widget, textual.widgets.OptionList):
        with contextlib.suppress(textual.widgets.option_list.OptionDoesNotExist):
            widget.replace_option_prompt(
                'loading-indicator',
                (str(widget.get_option('loading-indicator').prompt) + '.').replace('....', ''),
            )
    elif isinstance(widget, textual.widgets.RichLog):
        with contextlib.suppress(IndexError):
            loading_message = str(widget.lines[0].text)
            widget.clear()
            widget.write((loading_message + '.').replace('....', ''))
    else:
        raise ValueError(f'Unsupported widget type: {type(widget)}')


LOADING_MESSAGE = '⏳ loading...'


def add_inline_loading_indicator(widget):
    '''
    Given a textual.widgets.OptionList or a textual.widgets.RichLog instance, add a loading
    indicator to it.
    '''
    if isinstance(widget, textual.widgets.OptionList):
        loading_option = textual.widgets.option_list.Option(LOADING_MESSAGE, id='loading-indicator')
        widget.add_option(loading_option)
        widget.highlighted = None
    elif isinstance(widget, textual.widgets.RichLog):
        widget.write(LOADING_MESSAGE)
    else:
        raise ValueError(f'Unsupported widget type: {type(widget)}')

    return widget.set_interval(
        LOADING_DOT_INTERVAL_SECONDS,
        functools.partial(update_inline_loading_indicator, widget),
    )



import logging
import os
import re

logger = logging.getLogger(__name__)


SECRET_NAME_PATTERN = re.compile(r'^\w+$')
DEFAULT_SECRETS_DIRECTORY = '/run/secrets'


def load_credential(hook_config, config, credential_parameters):
    '''
    Given the hook configuration dict, the configuration dict, and a credential parameters tuple
    containing a secret name to load, read the secret from the corresponding container secrets file
    and return it.

    Raise ValueError if the credential parameters is not one element, the secret name is invalid, or
    the secret file cannot be read.
    '''
    try:
        (secret_name,) = credential_parameters
    except ValueError:
        name = ' '.join(credential_parameters)

        raise ValueError(f'Cannot load invalid secret name: "{name}"')

    if not SECRET_NAME_PATTERN.match(secret_name):
        raise ValueError(f'Cannot load invalid secret name: "{secret_name}"')

    try:
        with open(
            os.path.join(
                config.get('working_directory', ''),
                (hook_config or {}).get('secrets_directory', DEFAULT_SECRETS_DIRECTORY),
                secret_name,
            ),
            encoding='utf-8',
        ) as secret_file:
            return secret_file.read().rstrip(os.linesep)
    except (FileNotFoundError, OSError) as error:
        logger.warning(error)

        raise ValueError(f'Cannot load secret "{secret_name}" from file: {error.filename}')



import functools
import logging
import os
import re
import shlex
import subprocess
import sys

import borgmatic.execute
import borgmatic.logger

logger = logging.getLogger(__name__)


SOFT_FAIL_EXIT_CODE = 75
BORG_PLACEHOLDER_NAMES = {
    'hostname',
    'fqdn',
    'reverse-fqdn',
    'now',
    'utcnow',
    'unixtime',
    'user',
    'pid',
    'borgversion',
    'borgmajor',
    'borgminor',
    'borgpatch',
}


VARIABLE_PATTERN = re.compile(r'(?P<left_escape>\\)?\{(?P<name>[\w]+)(?P<right_escape>\\)?\}')


def resolve_variable(match, context, hook_description):
    '''
    Given a re.Match instance of VARIABLE_PATTERN representing a matched variable name to be
    interpolated, a context dict, and a description of the current command hook, lookup the matched
    variable name within the given context and return its value.

    If the match is escaped with backslashes, then instead of resolving the variable's value, strip
    off the backslashing and return the literal value.

    If the variable name isn't found in the given context (and isn't a Borg placeholder), then
    warn and return the literal value.
    '''
    name = match.group('name')

    # The would-be variable is escaped, so strip off the escaping and return the result without
    # resolving the name.
    if match.group('left_escape') and match.group('right_escape'):
        return '{' + name + '}'

    value = context.get(name)

    # The matched variable is in the context, so return its value.
    if value is not None:
        return shlex.quote(str(value))

    # The matched variable name isn't in the context. Warn about variables unknown to borgmatic, but
    # don't warn if the variable name happens to be a Borg placeholder, as Borg should hopefully
    # consume it.
    if name not in BORG_PLACEHOLDER_NAMES:
        logger.warning(
            f'Variable "{name}" is not supported in the {hook_description} hook',
        )

    # Return the whole string unaltered.
    return match.group(0)


def interpolate_context(hook_description, command, context):
    '''
    Given a config filename, a hook description, a single hook command, and a dict of context
    names/values, interpolate the values by "{name}" into the command and return the result.
    '''
    return VARIABLE_PATTERN.sub(
        functools.partial(resolve_variable, context=context, hook_description=hook_description),
        command,
    )


def make_environment(current_environment, sys_module=sys):
    '''
    Given the existing system environment as a map from environment variable name to value, return a
    copy of it, augmented with any extra environment variables that should be used when running
    command hooks.
    '''
    environment = dict(current_environment)

    # Detect whether we're running within a PyInstaller bundle. If so, set or clear LD_LIBRARY_PATH
    # based on the value of LD_LIBRARY_PATH_ORIG. This prevents library version information errors.
    if getattr(sys_module, 'frozen', False) and hasattr(sys_module, '_MEIPASS'):
        environment['LD_LIBRARY_PATH'] = environment.get('LD_LIBRARY_PATH_ORIG', '')

    return environment


def filter_hooks(command_hooks, before=None, after=None, action_names=None, state_names=None):
    '''
    Given a sequence of command hook dicts from configuration and one or more filters (before name,
    after name, a sequence of action names, and/or a sequence of execution result state names),
    filter down the command hooks to just the ones that match the given filters.
    '''
    return tuple(
        hook_config
        for hook_config in command_hooks or ()
        for config_action_names in (hook_config.get('when'),)
        for config_state_names in (hook_config.get('states'),)
        if before is None or hook_config.get('before') == before
        if after is None or hook_config.get('after') == after
        if action_names is None
        or config_action_names is None
        or set(config_action_names or ()).intersection(set(action_names))
        if state_names is None
        or config_state_names is None
        or set(config_state_names or ()).intersection(set(state_names))
    )


def execute_hooks(command_hooks, umask, working_directory, dry_run, **context):  # noqa: PLR0912
    '''
    Given a sequence of command hook dicts from configuration, a umask to execute with (or None), a
    working directory to execute with, and whether this is a dry run, run the commands for each
    hook. Or don't run them if this is a dry run.

    The context contains optional values interpolated by name into the hook commands.

    Raise ValueError if the umask cannot be parsed or a hook is invalid.
    Raise subprocesses.CalledProcessError if an error occurs in a hook.
    '''
    borgmatic.logger.add_custom_log_levels()

    dry_run_label = ' (dry run; not actually running hooks)' if dry_run else ''

    for hook_config in command_hooks:
        commands = hook_config.get('run')
        when_description = (
            f"{'/'.join(hook_config.get('when'))} " if hook_config.get('when') else ''
        )

        if 'before' in hook_config:
            description = f'before {when_description}{hook_config.get("before")}'
        elif 'after' in hook_config:
            description = f'after {when_description}{hook_config.get("after")}'
        else:
            raise ValueError(f'Invalid hook configuration: {hook_config}')

        if not commands:
            logger.debug(f'No commands to run for {description} hook')
            continue

        commands = [interpolate_context(description, command, context) for command in commands]

        if len(commands) == 1:
            logger.info(f'Running {description} command hook{dry_run_label}')
        else:
            logger.info(
                f'Running {len(commands)} commands for {description} hook{dry_run_label}',
            )

        if umask:
            parsed_umask = int(str(umask), 8)
            logger.debug(f'Setting hook umask to {oct(parsed_umask)}')
            original_umask = os.umask(parsed_umask)
        else:
            original_umask = None

        try:
            for command in commands:
                if dry_run:
                    continue

                borgmatic.execute.execute_command(  # noqa: S604
                    [command],
                    output_log_level=(
                        logging.ERROR if hook_config.get('after') == 'error' else logging.ANSWER
                    ),
                    shell=True,
                    environment=make_environment(os.environ),
                    working_directory=working_directory,
                    close_fds=True,
                )
        finally:
            if original_umask:
                os.umask(original_umask)


class Before_after_hooks:
    '''
    A Python context manager for executing command hooks both before and after the wrapped code.

    Example use as a context manager:

       with borgmatic.hooks.command.Before_after_hooks(
           command_hooks=config.get('commands'),
           before_after='do_stuff',
           umask=config.get('umask'),
           dry_run=dry_run,
           action_names=['create'],
       ):
            do()
            some()
            stuff()

    With that context manager in place, "before" command hooks execute before the wrapped code runs,
    and "after" command hooks execute after the wrapped code completes.
    '''

    def __init__(
        self,
        command_hooks,
        before_after,
        umask,
        working_directory,
        dry_run,
        action_names=None,
        **context,
    ):
        '''
        Given a sequence of command hook configuration dicts, the before/after name, a umask to run
        commands with, a working directory to run commands with, a dry run flag, a sequence of
        action names, and any context for the executed commands, save those data points for use
        below.
        '''
        self.command_hooks = command_hooks
        self.before_after = before_after
        self.umask = umask
        self.working_directory = working_directory
        self.dry_run = dry_run
        self.action_names = action_names
        self.context = context

    def __enter__(self):
        '''
        Run the configured "before" command hooks that match the initialized data points.
        '''
        try:
            execute_hooks(
                borgmatic.hooks.command.filter_hooks(
                    self.command_hooks,
                    before=self.before_after,
                    action_names=self.action_names,
                ),
                self.umask,
                self.working_directory,
                self.dry_run,
                **self.context,
            )
        except (OSError, subprocess.CalledProcessError) as error:
            if considered_soft_failure(error):
                raise

            # Trigger the after hook manually, since raising here will prevent it from being run
            # otherwise.
            self.__exit__(exception_type=type(error), exception=error, traceback=None)

            raise ValueError(f'Error running before {self.before_after} hook: {error}')

    def __exit__(self, exception_type, exception, traceback):
        '''
        Run the configured "after" command hooks that match the initialized data points.
        '''
        try:
            execute_hooks(
                borgmatic.hooks.command.filter_hooks(
                    self.command_hooks,
                    after=self.before_after,
                    action_names=self.action_names,
                    state_names=['fail' if exception_type else 'finish'],
                ),
                self.umask,
                self.working_directory,
                self.dry_run,
                **self.context,
            )
        except (OSError, subprocess.CalledProcessError) as error:
            if considered_soft_failure(error):
                raise

            raise ValueError(f'Error running after {self.before_after} hook: {error}')


@functools.cache
def considered_soft_failure(error):
    '''
    Given a configuration filename and an exception object, return whether the exception object
    represents a subprocess.CalledProcessError with a return code of SOFT_FAIL_EXIT_CODE. If so,
    that indicates that the error is a "soft failure", and should not result in an error.

    The results of this function are cached so that it can be called multiple times without logging
    multiple times.
    '''
    exit_code = getattr(error, 'returncode', None)
    if exit_code is None:
        return False

    if exit_code == SOFT_FAIL_EXIT_CODE:
        logger.info(
            f'Command hook exited with soft failure exit code ({SOFT_FAIL_EXIT_CODE}); skipping remaining repository actions',
        )
        return True

    return False



import logging
import shlex

import borgmatic.config.paths
import borgmatic.logger
from borgmatic.borg import environment, feature, flags
from borgmatic.execute import execute_command, execute_command_and_capture_output

logger = logging.getLogger(__name__)


def display_repository_info(
    repository_path,
    config,
    local_borg_version,
    repo_info_arguments,
    global_arguments,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a local or remote repository path, a configuration dict, the local Borg version, the
    arguments to the repo_info action, and global arguments as an argparse.Namespace, display summary
    information for the Borg repository or return JSON summary information.
    '''
    borgmatic.logger.add_custom_log_levels()
    lock_wait = config.get('lock_wait')
    extra_borg_options = config.get('extra_borg_options', {}).get(
        'repo_info' if feature.available(feature.Feature.REPO_INFO, local_borg_version) else 'info',
        '',
    )

    full_command = (
        (local_path,)
        + (
            ('repo-info',)
            if feature.available(feature.Feature.REPO_INFO, local_borg_version)
            else ('info',)
        )
        + (
            ('--info',)
            if logger.getEffectiveLevel() == logging.INFO and not repo_info_arguments.json
            else ()
        )
        + (
            ('--debug', '--show-rc')
            if logger.isEnabledFor(logging.DEBUG) and not repo_info_arguments.json
            else ()
        )
        + flags.make_flags('remote-path', remote_path)
        + flags.make_flags('umask', config.get('umask'))
        + flags.make_flags('lock-wait', lock_wait)
        + ('--log-json',)
        + (('--json',) if repo_info_arguments.json else ())
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + flags.make_repository_flags(repository_path, local_borg_version)
    )

    working_directory = borgmatic.config.paths.get_working_directory(config)
    borg_exit_codes = config.get('borg_exit_codes')

    if repo_info_arguments.json:
        return '\n'.join(
            execute_command_and_capture_output(
                full_command,
                environment=environment.make_environment(config),
                working_directory=working_directory,
                borg_local_path=local_path,
                borg_exit_codes=borg_exit_codes,
            )
        )

    execute_command(
        full_command,
        output_log_level=logging.ANSWER,
        environment=environment.make_environment(config),
        working_directory=working_directory,
        borg_local_path=local_path,
        borg_exit_codes=borg_exit_codes,
    )

    return None



import os

import borgmatic.borg.passcommand
import borgmatic.hooks.credential.parse

OPTION_TO_ENVIRONMENT_VARIABLE = {
    'archive_hostname': 'BORG_HOSTNAME',
    'archive_username': 'BORG_USERNAME',
    'borg_base_directory': 'BORG_BASE_DIR',
    'borg_config_directory': 'BORG_CONFIG_DIR',
    'borg_cache_directory': 'BORG_CACHE_DIR',
    'borg_files_cache_ttl': 'BORG_FILES_CACHE_TTL',
    'borg_security_directory': 'BORG_SECURITY_DIR',
    'borg_keys_directory': 'BORG_KEYS_DIR',
    'borg_key_file': 'BORG_KEY_FILE',
    'ssh_command': 'BORG_RSH',
    'temporary_directory': 'TMPDIR',
}

DEFAULT_BOOL_OPTION_TO_UNCONDITIONAL_ENVIRONMENT_VARIABLE = {
    'check_i_know_what_i_am_doing': 'BORG_CHECK_I_KNOW_WHAT_I_AM_DOING',
}

DEFAULT_BOOL_OPTION_TO_ENVIRONMENT_VARIABLE = {
    'debug_passphrase': 'BORG_DEBUG_PASSPHRASE',
    'display_passphrase': 'BORG_DISPLAY_PASSPHRASE',
    'relocated_repo_access_is_ok': 'BORG_RELOCATED_REPO_ACCESS_IS_OK',
    'unknown_unencrypted_repo_access_is_ok': 'BORG_UNKNOWN_UNENCRYPTED_REPO_ACCESS_IS_OK',
    'use_chunks_archive': 'BORG_USE_CHUNKS_ARCHIVE',
    'msgpack_version_check': 'BORG_MSGPACK_VERSION_CHECK',
}


def make_environment(config):
    '''
    Given a borgmatic configuration dict, convert it to a Borg environment variable dict, merge it
    with a copy of the current environment variables, and return the result.

    Do not reuse this environment across multiple Borg invocations, because it can include
    references to resources like anonymous pipes for passphrases—which can only be consumed once.

    Here's how native Borg precedence works for a few of the environment variables:

      1. BORG_PASSPHRASE, if set, is used first.
      2. BORG_PASSCOMMAND is used only if BORG_PASSPHRASE isn't set.
      3. BORG_PASSPHRASE_FD is used only if neither of the above are set.

    In borgmatic, we want to simulate this precedence order, but there are some additional
    complications. First, values can come from either configuration or from environment variables
    set outside borgmatic; configured options should take precedence. Second, when borgmatic gets a
    passphrase—directly from configuration or indirectly via a credential hook or a passcommand—we
    want to pass that passphrase to Borg via an anonymous pipe (+ BORG_PASSPHRASE_FD), since that's
    more secure than using an environment variable (BORG_PASSPHRASE).
    '''
    environment = dict(os.environ)

    for option_name, environment_variable_name in OPTION_TO_ENVIRONMENT_VARIABLE.items():
        value = config.get(option_name)

        if value is not None:
            environment[environment_variable_name] = str(value)

    if 'encryption_passphrase' in config:
        environment.pop('BORG_PASSPHRASE', None)
        environment.pop('BORG_PASSCOMMAND', None)

    if 'encryption_passcommand' in config:
        environment.pop('BORG_PASSCOMMAND', None)

    passphrase = borgmatic.hooks.credential.parse.resolve_credential(
        config.get('encryption_passphrase'),
        config,
    )

    if passphrase is None:
        passphrase = borgmatic.borg.passcommand.get_passphrase_from_passcommand(config)

    # If there's a passphrase (from configuration, from a configured credential, or from a
    # configured passcommand), send it to Borg via an anonymous pipe.
    if passphrase is not None:
        read_file_descriptor, write_file_descriptor = os.pipe()
        os.write(write_file_descriptor, passphrase.encode('utf-8'))
        os.close(write_file_descriptor)

        # This plus subprocess.Popen(..., close_fds=False) in execute.py is necessary for the Borg
        # child process to inherit the file descriptor.
        os.set_inheritable(read_file_descriptor, True)
        environment['BORG_PASSPHRASE_FD'] = str(read_file_descriptor)

    for (
        option_name,
        environment_variable_name,
    ) in DEFAULT_BOOL_OPTION_TO_ENVIRONMENT_VARIABLE.items():
        if os.environ.get(environment_variable_name) is None:
            value = config.get(option_name)
            if value is not None:
                environment[environment_variable_name] = 'YES' if value else 'NO'

    for (
        option_name,
        environment_variable_name,
    ) in DEFAULT_BOOL_OPTION_TO_UNCONDITIONAL_ENVIRONMENT_VARIABLE.items():
        value = config.get(option_name)
        if value is not None:
            environment[environment_variable_name] = 'YES' if value else 'NO'

    # On Borg 1.4.0a1+, take advantage of more specific exit codes. No effect on
    # older versions of Borg.
    environment['BORG_EXIT_CODES'] = 'modern'

    return environment



import argparse
import json
import logging
import shlex
import subprocess

import borgmatic.config.paths
import borgmatic.logger
from borgmatic.borg import environment, feature, flags, repo_info
from borgmatic.execute import DO_NOT_CAPTURE, execute_command

logger = logging.getLogger(__name__)


REPO_INFO_REPOSITORY_NOT_FOUND_EXIT_CODES = {2, 13, 15}


def create_repository(
    dry_run,
    repository_path,
    config,
    local_borg_version,
    global_arguments,
    encryption_mode,
    id_hash=None,
    key_location=None,
    source_repository=None,
    from_borg1=False,
    copy_crypt_key=False,
    append_only=None,
    storage_quota=None,
    make_parent_directories=False,
    local_path='borg',
    remote_path=None,
):
    '''
    Given a dry-run flag, a local or remote repository path, a configuration dict, the local Borg
    version, a Borg encryption mode, a Borg ID hash function name, a key location ("repokey" or
    "keyfile"), the path to another repo whose key material should be reused, whether that source
    repository is a Borg 1 repository, whether to copy the crypt key from the source repository,
    whether the repository should be append-only, the storage quota to use, whether to create any
    missing parent directories for the repository directory, the local Borg path, and the remote
    Borg path, create the repository. If the repository already exists, then log and skip creation.

    Raise ValueError if the requested encryption mode does not match that of the repository.
    Raise json.decoder.JSONDecodeError if the "borg info" JSON outputcannot be decoded.
    Raise subprocess.CalledProcessError if "borg info" returns an error exit code.
    '''
    try:
        # Suppress Borg's "repository does not exist" error log, so the user isn't confused by
        # seeing an error during successful repository creation.
        with borgmatic.logger.Logs_suppressed(msgid='Repository.DoesNotExist'):
            info_data = json.loads(
                repo_info.display_repository_info(
                    repository_path,
                    config,
                    local_borg_version,
                    argparse.Namespace(json=True),
                    global_arguments,
                    local_path,
                    remote_path,
                ),
            )

        encryption_data = info_data.get('encryption', {})
        repository_encryption_mode = encryption_data.get('mode', encryption_data.get('encryption'))
        repository_id_hash = encryption_data.get('id_hash')

        if encryption_mode != repository_encryption_mode:
            raise ValueError(
                f'Requested encryption mode "{encryption_mode}" does not match existing repository encryption mode "{repository_encryption_mode}"',
            )

        if id_hash and id_hash != repository_id_hash:
            raise ValueError(
                f'Requested ID hash function "{id_hash}" does not match existing repository ID hash function "{repository_id_hash}"',
            )

        logger.info('Repository already exists. Skipping creation.')
        return
    except subprocess.CalledProcessError as error:
        if error.returncode not in REPO_INFO_REPOSITORY_NOT_FOUND_EXIT_CODES:
            raise

    lock_wait = config.get('lock_wait')
    umask = config.get('umask')
    extra_borg_options_from_init = config.get('extra_borg_options', {}).get('init', '')
    extra_borg_options = config.get('extra_borg_options', {}).get('repo-create', '')

    if extra_borg_options_from_init:
        logger.warning(
            'The "init" option in "extra_borg_options" is deprecated and will be removed from a future release; use "repo_create" instead.'
        )

    repo_create_command = (
        (local_path,)
        + (
            ('repo-create',)
            if feature.available(feature.Feature.REPO_CREATE, local_borg_version)
            else ('init',)
        )
        + (('--encryption', encryption_mode) if encryption_mode else ())
        + (('--id-hash', id_hash) if id_hash else ())
        + (('--key-location', key_location) if key_location else ())
        + (('--other-repo', source_repository) if source_repository else ())
        + (('--from-borg1',) if from_borg1 else ())
        + (('--copy-crypt-key',) if copy_crypt_key else ())
        + (('--append-only',) if append_only else ())
        + (('--storage-quota', storage_quota) if storage_quota else ())
        + (('--make-parent-dirs',) if make_parent_directories else ())
        + (('--info',) if logger.getEffectiveLevel() == logging.INFO else ())
        + (('--debug',) if logger.isEnabledFor(logging.DEBUG) else ())
        + (('--lock-wait', str(lock_wait)) if lock_wait else ())
        + (('--remote-path', remote_path) if remote_path else ())
        + (('--umask', str(umask)) if umask else ())
        + (tuple(shlex.split(extra_borg_options)) if extra_borg_options else ())
        + (tuple(shlex.split(extra_borg_options_from_init)) if extra_borg_options_from_init else ())
        + flags.make_repository_flags(repository_path, local_borg_version)
    )

    if dry_run:
        logger.info('Skipping repository creation (dry run)')
        return

    # Do not capture output here, so as to support interactive prompts.
    execute_command(
        repo_create_command,
        output_file=DO_NOT_CAPTURE,
        environment=environment.make_environment(config),
        working_directory=borgmatic.config.paths.get_working_directory(config),
        borg_local_path=local_path,
        borg_exit_codes=config.get('borg_exit_codes'),
    )



import logging
import operator

import borgmatic.hooks.credential.parse
import borgmatic.hooks.monitoring.logs
import borgmatic.hooks.monitoring.monitor

logger = logging.getLogger(__name__)


DEFAULT_LOGS_SIZE_LIMIT_BYTES = 100000
HANDLER_IDENTIFIER = 'apprise'


def initialize_monitor(hook_config, config, config_filename, monitoring_log_level, dry_run):
    '''
    Add a handler to the root logger that stores in memory the most recent logs emitted. That way,
    we can send them all to an Apprise notification service upon a finish or failure state. But skip
    this if the "send_logs" option is false.
    '''
    if hook_config.get('send_logs', False) is False:
        return

    logs_size_limit = max(
        hook_config.get('logs_size_limit', DEFAULT_LOGS_SIZE_LIMIT_BYTES)
        - len(borgmatic.hooks.monitoring.logs.PAYLOAD_TRUNCATION_INDICATOR),
        0,
    )

    borgmatic.hooks.monitoring.logs.add_handler(
        borgmatic.hooks.monitoring.logs.Forgetful_buffering_handler(
            HANDLER_IDENTIFIER,
            logs_size_limit,
            monitoring_log_level,
        ),
    )


def ping_monitor(hook_config, config, config_filename, state, monitoring_log_level, dry_run):
    '''
    Ping the configured Apprise service URLs. Use the given configuration filename in any log
    entries. If this is a dry run, then don't actually ping anything.
    '''
    try:
        import apprise  # noqa: PLC0415
        from apprise import NotifyFormat, NotifyType  # noqa: PLC0415
    except ImportError:  # pragma: no cover
        logger.warning(
            'Unable to import Apprise in its monitoring hook; try installing "borgmatic[Apprise]"'
        )
        return

    state_to_notify_type = {
        'start': NotifyType.INFO,
        'finish': NotifyType.SUCCESS,
        'fail': NotifyType.FAILURE,
        'log': NotifyType.INFO,
    }

    run_states = hook_config.get('states', ['fail'])

    if state.name.lower() not in run_states:
        return

    state_config = hook_config.get(
        state.name.lower(),
        {
            'title': f'A borgmatic {state.name} event happened',
            'body': f'A borgmatic {state.name} event happened',
        },
    )

    if not hook_config.get('services'):
        logger.info('No Apprise services to ping')
        return

    dry_run_string = ' (dry run; not actually pinging)' if dry_run else ''
    labels_string = ', '.join(map(operator.itemgetter('label'), hook_config.get('services')))
    logger.info(f'Pinging Apprise services: {labels_string}{dry_run_string}')

    apprise_object = apprise.Apprise()
    apprise_object.add(
        [
            borgmatic.hooks.credential.parse.resolve_credential(service['url'], config)
            for service in hook_config.get('services')
        ]
    )

    if dry_run:
        return

    body = state_config.get('body')

    if state in {
        borgmatic.hooks.monitoring.monitor.State.FINISH,
        borgmatic.hooks.monitoring.monitor.State.FAIL,
        borgmatic.hooks.monitoring.monitor.State.LOG,
    }:
        formatted_logs = borgmatic.hooks.monitoring.logs.format_buffered_logs_for_payload(
            HANDLER_IDENTIFIER,
        )
        if formatted_logs:
            body += f'\n\n{formatted_logs}'

    result = apprise_object.notify(
        title=state_config.get('title', ''),
        body=body,
        body_format=NotifyFormat.TEXT,
        notify_type=state_to_notify_type[state.name.lower()],
    )

    if result is False:
        logger.warning('Error sending some Apprise notifications')


def destroy_monitor(hook_config, config, monitoring_log_level, dry_run):
    '''
    Remove the monitor handler that was added to the root logger. This prevents the handler from
    getting reused by other instances of this monitor.
    '''
    borgmatic.hooks.monitoring.logs.remove_handler(HANDLER_IDENTIFIER)



import json
import logging
import shutil
import subprocess

import borgmatic.borg.pattern
from borgmatic.execute import execute_command_and_capture_output

IS_A_HOOK = False

logger = logging.getLogger(__name__)


def resolve_database_option(option, data_source, connection_params=None, restore=False):
    '''
    Resolves a database option from the given data source configuration dict and
    connection parameters dict. If restore is set to True it will consider the
    `restore_<option>` instead.

    Returns the resolved option or None. Can raise a ValueError if the hostname lookup
    results in a container IP check.
    '''
    # Special case `hostname` since it overlaps with `container`
    if option == 'hostname':
        return get_hostname_from_config(data_source, connection_params, restore)
    if connection_params and (value := connection_params.get(option)):
        return value
    if restore and f'restore_{option}' in data_source:
        return data_source[f'restore_{option}']

    return data_source.get(option)


def get_hostname_from_config(data_source, connection_params=None, restore=False):
    '''
    Specialisation of `resolve_database_option` to handle the extra complexity of
    the hostname option to also handle containers.

    Returns a hostname/IP or raises an ValueError if a container IP lookup fails.
    '''
    # connection params win, full stop
    if connection_params:
        if container := connection_params.get('container'):
            return get_ip_from_container(container)
        if hostname := connection_params.get('hostname'):
            return hostname
    # ... then try the restore config
    if restore:
        if 'restore_container' in data_source:
            return get_ip_from_container(data_source['restore_container'])
        if 'restore_hostname' in data_source:
            return data_source['restore_hostname']
    # ... and finally fall back to the normal options
    if 'container' in data_source:
        return get_ip_from_container(data_source['container'])

    return data_source.get('hostname')


def get_ip_from_container(container):
    '''
    Determine the IP for a given container name via podman and docker.

    Returns an IP or raises a ValueError if the lookup fails.
    '''
    engines = (shutil.which(engine) for engine in ('docker', 'podman'))
    engines = [engine for engine in engines if engine]

    if not engines:
        raise ValueError("Neither 'docker' nor 'podman' could be found on the system")

    last_error = None
    for engine in engines:
        try:
            output = '\n'.join(
                execute_command_and_capture_output(
                    (
                        engine,
                        'container',
                        'inspect',
                        '--format={{json .NetworkSettings}}',
                        container,
                    )
                )
            )
        except subprocess.CalledProcessError as error:
            last_error = error
            logger.debug(f"Could not find container '{container}' with engine '{engine}'")
            continue  # Container does not exist

        try:
            network_data = json.loads(output.strip())
        except json.JSONDecodeError as e:
            raise ValueError(f'Could not decode JSON output from {engine}') from e
        if main_ip := network_data.get('IPAddress'):
            return main_ip
        # No main IP found, try the networks
        for network in network_data.get('Networks', {}).values():
            if ip := network.get('IPAddress'):
                return ip

    if last_error:
        raise last_error

    raise ValueError(
        f"Could not determine ip address for container '{container}'; running in host mode or userspace networking?"
    )


def inject_pattern(patterns, data_source_pattern, override_excludes=True):
    '''
    Given a list of borgmatic.borg.pattern.Pattern instances representing the configured patterns,
    insert the given data source pattern at the start of the list. The idea is that borgmatic is
    injecting its own custom pattern specific to a data source hook into the user's configured
    patterns so that the hook's data gets included in the backup.

    As part of this injection, if the data source pattern is a root pattern and override_excludes is
    True, also insert an "include" version of the given root pattern, in an attempt to preempt any
    of the user's configured exclude patterns that may follow. The is to support use cases like
    borgmatic injecting its own patterns for things like database dumps or bootstrap metadata, where
    we don't want them to get accidentally excluded.
    '''
    if data_source_pattern.type == borgmatic.borg.pattern.Pattern_type.ROOT and override_excludes:
        patterns.insert(
            0,
            borgmatic.borg.pattern.Pattern(
                path=data_source_pattern.path,
                type=borgmatic.borg.pattern.Pattern_type.INCLUDE,
                style=data_source_pattern.style,
                device=data_source_pattern.device,
                source=borgmatic.borg.pattern.Pattern_source.HOOK,
            ),
        )

    patterns.insert(0, data_source_pattern)


def replace_pattern(patterns, pattern_to_replace, data_source_pattern):
    '''
    Given a list of borgmatic.borg.pattern.Pattern instances representing the configured patterns,
    replace the given pattern with the given data source pattern. The idea is that borgmatic is
    replacing a configured pattern with its own modified pattern specific to a data source hook so
    that the hook's data gets included in the backup.

    If the pattern to replace can't be found in the given patterns, then just inject the data source
    pattern at the start of the list.
    '''
    try:
        index = patterns.index(pattern_to_replace)
    except ValueError:
        inject_pattern(patterns, data_source_pattern, override_excludes=False)

        return

    patterns[index] = data_source_pattern



import os

import textual.binding
import textual.containers

import borgmatic.actions.browse.archive
import borgmatic.actions.browse.archives_list
import borgmatic.actions.browse.configuration_files_list
import borgmatic.actions.browse.directory_list
import borgmatic.actions.browse.file_preview
import borgmatic.actions.browse.icons
import borgmatic.actions.browse.repositories_list


def make_next_panel(focused_panel, option_id):
    '''
    Given a focused panel widget and the selected option ID, return the next panel corresponding to
    that selection. This is the mechanism by which the user can successively drill down from
    configuration file to repository to archive to root directory to non-root directory or file.

    If the particular option ID on the focused panel doesn't have a supported next panel, then
    return None.
    '''
    if isinstance(
        focused_panel, borgmatic.actions.browse.configuration_files_list.Configuration_files_list
    ):
        return borgmatic.actions.browse.repositories_list.Repositories_list(
            config=focused_panel.configs[option_id]
        )

    if isinstance(focused_panel, borgmatic.actions.browse.repositories_list.Repositories_list):
        return borgmatic.actions.browse.archives_list.Archives_list(
            config=focused_panel.config, repository=focused_panel.repositories[option_id]
        )

    if isinstance(focused_panel, borgmatic.actions.browse.archives_list.Archives_list):
        return borgmatic.actions.browse.directory_list.Directory_list(
            config=focused_panel.config, repository=focused_panel.repository, archive_name=option_id
        )

    if isinstance(focused_panel, borgmatic.actions.browse.directory_list.Directory_list):
        option = focused_panel.get_option(option_id)

        if option.prompt.startswith(
            borgmatic.actions.browse.icons.PATH_TYPE_ICONS[
                borgmatic.actions.browse.archive.Path_type.DIRECTORY.value
            ]
        ):
            return borgmatic.actions.browse.directory_list.Directory_list(
                focused_panel.config,
                focused_panel.repository,
                focused_panel.archive_name,
                path_loaded=focused_panel.path_loaded,
                path_components=(*focused_panel.path_components, option_id),
            )

        if option.prompt.startswith(
            borgmatic.actions.browse.icons.PATH_TYPE_ICONS[
                borgmatic.actions.browse.archive.Path_type.FILE.value
            ]
        ):
            return borgmatic.actions.browse.file_preview.File_preview(
                focused_panel.config,
                focused_panel.repository,
                focused_panel.archive_name,
                file_path=os.path.sep.join((*focused_panel.path_components, option_id)),
            )

    return None


class Carousel(textual.containers.Horizontal):
    BINDINGS = (
        textual.binding.Binding(
            key='left,h', action='previous', description='previous', priority=True
        ),
    )

    def __init__(self, panels):
        self.panels = panels
        self.focused_panel = panels[0]

        super().__init__()

    def compose(self):
        '''
        Compose with each of the contained panels and focus the first one.
        '''
        yield from self.panels

        self.focused_panel.focus()

    def action_previous(self):
        '''
        Make the previous panel into the focused panel.
        '''
        previous_panel_index = self.panels.index(self.focused_panel) - 1

        if previous_panel_index < 0:
            return

        self.focused_panel.styles.display = 'none'

        self.focused_panel = self.panels[previous_panel_index]
        self.focused_panel.styles.display = 'block'
        self.focused_panel.focus()

    def action_next(self, option_id):
        '''
        Hide the current focused panel and create the next one.
        '''
        next_panel_index = self.panels.index(self.focused_panel) + 1

        if next_panel_index < len(self.panels):
            next_panel = self.panels[next_panel_index]
            next_panel.styles.display = 'block'
        else:
            next_panel = make_next_panel(self.focused_panel, option_id)

            if next_panel is None:
                self.notify('Cannot display this content', severity='warning')
                return

            self.panels.append(next_panel)
            next_panel.highlighted = 0

        self.focused_panel.styles.display = 'none'
        self.focused_panel = next_panel
        self.focused_panel.focus()
        self.mount(self.focused_panel)

    def on_option_list_option_highlighted(self, event):
        '''
        The highlighted option has changed, so truncate any next panels.
        '''
        next_panel_index = self.panels.index(self.focused_panel) + 1

        del self.panels[next_panel_index:]

    def on_option_list_option_selected(self, event):
        '''
        An option has been selected, so advance to the next panel—unless the option selected is
        "..", in which case go to the previous panel.
        '''
        if (
            event.option_list != self.focused_panel or event.option_id == 'loading-indicator'
        ):  # pragma: no cover
            return

        if event.option_id == '..':
            self.action_previous()
        else:
            self.action_next(event.option_id)



import logging

import borgmatic.borg.repo_create

logger = logging.getLogger(__name__)


def run_repo_create(
    repository,
    config,
    local_borg_version,
    repo_create_arguments,
    global_arguments,
    local_path,
    remote_path,
):
    '''
    Run the "repo-create" action for the given repository.
    '''
    logger.info('Creating repository')

    encryption_mode = repo_create_arguments.encryption_mode or repository.get('encryption')

    if not encryption_mode:
        raise ValueError(
            'With the repo-create action, either the --encryption flag or the repository encryption option is required.',
        )

    borgmatic.borg.repo_create.create_repository(
        dry_run=global_arguments.dry_run,
        repository_path=repository['path'],
        config=config,
        local_borg_version=local_borg_version,
        global_arguments=global_arguments,
        encryption_mode=encryption_mode,
        id_hash=(
            repository.get('id_hash')
            if repo_create_arguments.id_hash is None
            else repo_create_arguments.id_hash
        ),
        key_location=(
            repository.get('key_location')
            if repo_create_arguments.key_location is None
            else repo_create_arguments.key_location
        ),
        source_repository=repo_create_arguments.source_repository,
        from_borg1=repo_create_arguments.from_borg1,
        copy_crypt_key=repo_create_arguments.copy_crypt_key,
        append_only=(
            repository.get('append_only')
            if repo_create_arguments.append_only is None
            else repo_create_arguments.append_only
        ),
        storage_quota=(
            repository.get('storage_quota')
            if repo_create_arguments.storage_quota is None
            else repo_create_arguments.storage_quota
        ),
        make_parent_directories=(
            repository.get('make_parent_directories')
            if repo_create_arguments.make_parent_directories is None
            else repo_create_arguments.make_parent_directories
        ),
        local_path=local_path,
        remote_path=remote_path,
    )



import borgmatic.hooks.dispatch


class Dump_cleanup:
    '''
    A Python context manager for removing data source dumps before and after wrapped code. Besides
    doing this for general filesystem cleanliness reasons, leaving old dumps lying around can
    prevent future dumping from working (in the case of filesystem snapshots) or cause Borg hangs
    (in the case of database dump named pipes).

    Example use as a context manager:

        with borgmatic.actions.dump.Dump_cleanup(
            config, borgmatic_runtime_directory, patterns, dry_run,
        ):
            do_something_like_perform_a_dump_or_restore()
    '''

    def __init__(self, config, borgmatic_runtime_directory, patterns, dry_run):
        '''
        Given a configuration dict, the borgmatic runtime directory, the configured patterns, and
        whether this is a dry-run, store these values for use below.
        '''
        self.config = config
        self.borgmatic_runtime_directory = borgmatic_runtime_directory
        self.patterns = patterns
        self.dry_run = dry_run

    def __enter__(self):
        '''
        Remove all data source dumps that exist prior to the wrapped code running.
        '''
        borgmatic.hooks.dispatch.call_hooks_even_if_unconfigured(
            'remove_data_source_dumps',
            self.config,
            borgmatic.hooks.dispatch.Hook_type.DATA_SOURCE,
            self.borgmatic_runtime_directory,
            self.patterns,
            self.dry_run,
        )

    def __exit__(self, exception_type, exception, traceback):
        '''
        Remove all data source dumps, including any created by the wrapped code.
        '''
        borgmatic.hooks.dispatch.call_hooks_even_if_unconfigured(
            'remove_data_source_dumps',
            self.config,
            borgmatic.hooks.dispatch.Hook_type.DATA_SOURCE,
            self.borgmatic_runtime_directory,
            self.patterns,
            self.dry_run,
        )



import logging

import borgmatic.actions.dump
import borgmatic.actions.json
import borgmatic.borg.create
import borgmatic.borg.feature
import borgmatic.borg.rename
import borgmatic.borg.repo_list
import borgmatic.config.paths
import borgmatic.hooks.dispatch
from borgmatic.actions import pattern

logger = logging.getLogger(__name__)


def run_create(
    config_filename,
    repository,
    config,
    config_paths,
    local_borg_version,
    create_arguments,
    global_arguments,
    dry_run_label,
    local_path,
    remote_path,
):
    '''
    Run the "create" action for the given repository.

    If create_arguments.json is True, yield the JSON output from creating the archive.
    '''
    if config.get('list_details') and config.get('progress'):
        raise ValueError(
            'With the create action, only one of --list/--files/list_details and --progress/progress can be used.',
        )

    if config.get('list_details') and create_arguments.json:
        raise ValueError(
            'With the create action, only one of --list/--files/list_details and --json can be used.',
        )

    logger.info(f'Creating archive{dry_run_label}')
    working_directory = borgmatic.config.paths.get_working_directory(config)

    with borgmatic.config.paths.Runtime_directory(config) as borgmatic_runtime_directory:
        patterns = pattern.process_patterns(
            pattern.collect_patterns(config, working_directory),
            config,
            working_directory,
            borgmatic_runtime_directory,
        )

        original_patterns = list(patterns)

        # Use the original patterns so as to disregard any modifications made by any data source
        # hooks, e.g. via dump_data_sources() below.
        with borgmatic.actions.dump.Dump_cleanup(
            config, borgmatic_runtime_directory, original_patterns, global_arguments.dry_run
        ):
            active_dumps = borgmatic.hooks.dispatch.call_hooks(
                'dump_data_sources',
                config,
                borgmatic.hooks.dispatch.Hook_type.DATA_SOURCE,
                config_paths,
                borgmatic_runtime_directory,
                patterns,
                global_arguments.dry_run,
            )

            # Process the patterns again in case any data source hooks updated them. Without this step,
            # we could end up with duplicate paths that cause Borg to hang when it tries to read from
            # the same named pipe twice.
            patterns = pattern.process_patterns(
                patterns,
                config,
                working_directory,
                borgmatic_runtime_directory,
                skip_expand_paths=config_paths,
            )
            stream_processes = [
                process for processes in active_dumps.values() for process in processes
            ]

            # If we have stream processes, we first create an archive with .checkpoint suffix. This is
            # to make sure we only create a real archive if all the streaming processes completed
            # successfully (create_archive will fail if a streaming process fails, but the archive might
            # have already been created at this point).
            use_checkpoint = bool(stream_processes)

            json_output = borgmatic.borg.create.create_archive(
                global_arguments.dry_run,
                repository['path'],
                config,
                patterns,
                local_borg_version,
                global_arguments,
                borgmatic_runtime_directory,
                archive_suffix='.checkpoint' if use_checkpoint else '',
                local_path=local_path,
                remote_path=remote_path,
                json=create_arguments.json,
                comment=create_arguments.comment,
                stream_processes=stream_processes,
            )

            if use_checkpoint:
                rename_checkpoint_archive(
                    repository['path'],
                    global_arguments,
                    config,
                    local_borg_version,
                    local_path,
                    remote_path,
                )

            if json_output:
                output = borgmatic.actions.json.parse_json(json_output, repository.get('label'))
                if use_checkpoint:
                    # Patch archive name and ID
                    renamed_archive = borgmatic.borg.repo_list.get_latest_archive(
                        repository['path'],
                        config,
                        local_borg_version,
                        global_arguments,
                        local_path,
                        remote_path,
                    )

                    output['archive']['name'] = renamed_archive['name']
                    output['archive']['id'] = renamed_archive['id']

                yield output


def rename_checkpoint_archive(
    repository_path,
    global_arguments,
    config,
    local_borg_version,
    local_path,
    remote_path,
):
    '''
    Renames the latest archive to not have a '.checkpoint' suffix.

    Raises ValueError if
    - there is not latest archive
    - the latest archive does not have a '.checkpoint' suffix

    Implementation note: We cannot reliably get the just created archive name.
    So we resort to listing the archives and picking the last one.

    A similar comment applies to retrieving the ID of the renamed archive.
    '''
    archive = borgmatic.borg.repo_list.get_latest_archive(
        repository_path,
        config,
        local_borg_version,
        global_arguments,
        local_path,
        remote_path,
        consider_checkpoints=True,
    )

    archive_name = archive['name']

    if not archive_name.endswith('.checkpoint'):
        raise ValueError(f'Latest archive did not have a .checkpoint suffix. Got: {archive_name}')

    new_archive_name = archive_name.removesuffix('.checkpoint')

    logger.info(f'Renaming archive {archive_name} -> {new_archive_name}')

    borgmatic.borg.rename.rename_archive(
        repository_path,
        (
            archive['id']
            if borgmatic.borg.feature.available(
                borgmatic.borg.feature.Feature.ARCHIVE_SERIES, local_borg_version
            )
            else archive['name']
        ),
        new_archive_name,
        global_arguments.dry_run,
        config,
        local_borg_version,
        local_path,
        remote_path,
    )



import json
import logging

logger = logging.getLogger(__name__)


def parse_json(borg_json_output, label):
    '''
    Given a Borg JSON output string, parse it as JSON into a dict. Inject the given borgmatic
    repository label into it and return the dict.

    Raise JSONDecodeError if the JSON output cannot be parsed.
    '''
    lines = borg_json_output.splitlines()
    start_line_index = 0

    # Scan forward to find the first line starting with "{" and assume that's where the JSON starts.
    for line_index, line in enumerate(lines):
        if line.startswith('{'):
            start_line_index = line_index
            break

    json_data = json.loads('\n'.join(lines[start_line_index:]))

    if 'repository' not in json_data:
        return json_data

    json_data['repository']['label'] = label or ''

    return json_data
